import numpy as np

from agox.models.descriptors import Fingerprint


def test_cache(environment_and_dataset):
    environment, dataset = environment_and_dataset

    descriptor = Fingerprint(environment=environment, use_cache=True)
    descriptor_no_cache = Fingerprint(environment=environment, use_cache=False)

    F_all_cache = descriptor.get_features(dataset)
    F_all_no_cache = descriptor_no_cache.get_features(dataset)

    np.testing.assert_allclose(F_all_cache, F_all_no_cache)

    # Use the cached descriptor again:
    F_all_cache_2 = descriptor.get_features(dataset)

    # Compare cached and non-cached:
    np.testing.assert_allclose(F_all_cache, F_all_cache_2)



import os

import numpy as np

from agox.databases import Database


def test_new_database(tmp_path, environment_and_dataset):
    # Make and move to temp dir:
    start_dir = os.getcwd()
    d = tmp_path / "database_test"
    os.mkdir(d)
    os.chdir(d)

    database_created = Database("db1.db")
    environment, dataset = environment_and_dataset

    meta_data_names = ["str", "int", "float", "1darr", "2darr"]
    meta_data_examples = ["str", 0, 0.1, np.array([0, 1, 2]), np.random.rand(3, 3)]

    for candidate in dataset:
        for meta_data_name, meta_data in zip(meta_data_names, meta_data_examples):
            candidate.add_meta_information(meta_data_name, meta_data)
        database_created.store_candidate(candidate)

    assert len(database_created) == len(dataset)

    for db_cand, data_cand in zip(database_created.get_all_candidates(), dataset):
        assert (db_cand.positions == data_cand.positions).all()
        assert (db_cand.cell == data_cand.cell).all()
        assert (db_cand.pbc == data_cand.pbc).all()
        for meta_data_name in meta_data_names:
            assert db_cand.get_meta_information(meta_data_name) is not None
            assert data_cand.get_meta_information(meta_data_name) is not None
            assert np.array(
                db_cand.get_meta_information(meta_data_name) == data_cand.get_meta_information(meta_data_name)
            ).all()

    database_created_energies = np.array(
        [atoms.get_potential_energy() for atoms in database_created.get_all_candidates()]
    )
    dataset_energies = np.array([atoms.get_potential_energy() for atoms in dataset])
    assert (database_created_energies == dataset_energies).all()

    database_created_forces = np.array([atoms.get_forces() for atoms in database_created.get_all_candidates()])
    dataset_forces = np.array([atoms.get_forces() for atoms in dataset])
    assert (database_created_forces == dataset_forces).all()
    database_created.write(already_stored=True, force_write=True)

    # Load the database from disk:
    database_loaded = Database("db1.db")
    database_loaded.restore_to_memory()

    assert len(database_loaded) == len(dataset)

    for db_cand, data_cand in zip(database_loaded.get_all_candidates(), dataset):
        assert (db_cand.positions == data_cand.positions).all()
        assert (db_cand.cell == data_cand.cell).all()
        assert (db_cand.pbc == data_cand.pbc).all()
        for meta_data_name in meta_data_names:
            print(meta_data_name)
            assert db_cand.get_meta_information(meta_data_name) is not None
            assert data_cand.get_meta_information(meta_data_name) is not None
            assert np.array(
                db_cand.get_meta_information(meta_data_name) == data_cand.get_meta_information(meta_data_name)
            ).all()

    database_loaded_energies = np.array(
        [atoms.get_potential_energy() for atoms in database_loaded.get_all_candidates()]
    )
    dataset_energies = np.array([atoms.get_potential_energy() for atoms in dataset])
    assert (database_loaded_energies == dataset_energies).all()

    database_loaded_forces = np.array([atoms.get_forces() for atoms in database_loaded.get_all_candidates()])
    dataset_forces = np.array([atoms.get_forces() for atoms in dataset])

    assert (database_loaded_forces == dataset_forces).all()
    database_loaded.write(already_stored=True, force_write=True)

    # Stuff about restoring parts of the database:
    energies_directly_restored = database_loaded.get_all_energies()
    assert (energies_directly_restored == dataset_energies).all()

    atoms_direct = database_loaded.get_structure(10)
    assert (atoms_direct.positions == dataset[9].positions).all()  # Index offset

    atoms_best_direct = database_loaded.get_best_structure()
    E = atoms_best_direct.get_potential_energy()
    assert E == np.min(dataset_energies)

    ebest = database_loaded.get_best_energy()
    assert E == ebest

    latest = database_loaded.get_most_recent_candidate()
    assert (latest.positions == dataset[-1].positions).all()

    # Move back to start dir as other tests fail otherwise.
    os.chdir(start_dir)



import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms
from ase.optimize import BFGS

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RattleGenerator
from agox.models.descriptors import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.postprocessors import RelaxPostprocess
from agox.samplers import MetropolisSampler

# Manually set seed and database-index
seed = 42
database_index = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# database_index = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au8Ni8",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=5)

##############################################################################
# Search Settings:
##############################################################################

kernel = C(1, (1, 100)) * RBF(20, (10, 30))
descriptor = SOAP(
    environment=environment,
    r_cut=5.0,
    nmax=3,
    lmax=2,
    sigma=1,
    weight=True,
    periodic=False,
)

model = SparseGPR(
    database=database,
    kernel=kernel,
    descriptor=descriptor,
    n_optimize=0,
    noise=0.01,  # eV/atom
    iteration_start_training=0,
    jitter=1e-4,  # only for CI testing
)

# Number of steps is very low - should be set higher for a real search!
relaxer = RelaxPostprocess(model, optimizer=BFGS, order=2, optimizer_run_kwargs={"steps": 5, "fmax": 0.1})

sampler = MetropolisSampler(temperature=0.25, order=4)

rattle_generator = RattleGenerator(**environment.get_confinement(), environment=environment, sampler=sampler, order=1)

# With the model pre-relax we dont want to take many steps in the real potential!
# As we are training the model all the data is saved with the store_trajectory argument.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    store_trajectory=True,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 3},
    order=3,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(rattle_generator, database, sampler, evaluator, relaxer, seed=seed)

agox.run(N_iterations=10)



import pytest

from agox.test.run_tests.run_utils import agox_test_run


@pytest.mark.ray
@pytest.mark.parametrize("mode", ["rex_emt"])
def test_replica_exchange(mode, tmp_path, cmd_options, ray_fix):
    agox_test_run(mode, tmp_path, cmd_options)



import os
import sys
from math import erf
from itertools import product
from scipy.spatial.distance import cdist

import time

import numpy as np
cimport numpy as np

from libc.math cimport *

from cymem.cymem cimport Pool
cimport cython

try:
    cwd = sys.argv[1]
except:
    cwd = os.getcwd()

# Custom functions
ctypedef struct Point:
    double coord[3]

cdef Point subtract(Point p1, Point p2):
    cdef Point p
    p.coord[0] = p1.coord[0] - p2.coord[0]
    p.coord[1] = p1.coord[1] - p2.coord[1]
    p.coord[2] = p1.coord[2] - p2.coord[2]
    return p

cdef Point add(Point p1, Point p2):
    cdef Point p
    p.coord[0] = p1.coord[0] + p2.coord[0]
    p.coord[1] = p1.coord[1] + p2.coord[1]
    p.coord[2] = p1.coord[2] + p2.coord[2]
    return p

cdef double norm(Point p):
    return sqrt(p.coord[0]*p.coord[0] + p.coord[1]*p.coord[1] + p.coord[2]*p.coord[2])

cdef double euclidean(Point p1, Point p2):
    return norm(subtract(p1,p2))

cdef double dot(Point v1, Point v2):
    return v1.coord[0]*v2.coord[0] + v1.coord[1]*v2.coord[1] + v1.coord[2]*v2.coord[2]

cdef double get_angle(Point v1, Point v2):
    """
    Returns angle with convention [0,pi]
    """
    norm1 = norm(v1)
    norm2 = norm(v2)
    arg = dot(v1,v2)/(norm1*norm2)
    # This is added to correct for numerical errors
    if arg < -1:
        arg = -1.
    elif arg > 1:
        arg = 1.
    return acos(arg)

@cython.cdivision(True)
cdef double f_cutoff(double r, double gamma, double Rcut):
    """
    Polinomial cutoff function in the, with the steepness determined by "gamma"
    gamma = 2 resembels the cosine cutoff function.
    For large gamma, the function goes towards a step function at Rc.
    """
    if not gamma == 0:
        return 1 + gamma*pow(r/Rcut, gamma+1) - (gamma+1)*pow(r/Rcut, gamma)
    else:
        return 1

@cython.cdivision(True)
cdef double f_cutoff_grad(double r, double gamma, double Rcut):
    if not gamma == 0:
        return gamma*(gamma+1)/Rcut * (pow(r/Rcut, gamma) - pow(r/Rcut, gamma-1))
    else:
        return 0

def convert_atom_types(num):
    cdef int Natoms = len(num)
    cdef list atomic_types = sorted(list(set(num)))
    cdef int Ntypes = len(atomic_types)
    cdef list num_converted = [0]*Natoms
    cdef int i, j
    for i in range(Natoms):
        for j in range(Ntypes):
            if num[i] == atomic_types[j]:
                num_converted[i] = j
    return num_converted

class Angular_Fingerprint(object):
    """ comparator for construction of angular fingerprints
    """
    def __init__(self, atoms, Rc1=4.0, Rc2=4.0, binwidth1=0.1, Nbins2=30, sigma1=0.2, sigma2=0.10, nsigma=4, eta=1, gamma=3, use_angular=True):
        """ Set a common cut of radius
        """

        self.Rc1 = Rc1
        self.Rc2 = Rc2
        self.binwidth1 = binwidth1
        self.Nbins2 = Nbins2
        self.binwidth2 = np.pi / Nbins2
        self.sigma1 = sigma1
        self.sigma2 = sigma2
        self.nsigma = nsigma
        self.eta = eta
        self.gamma = gamma
        self.use_angular = use_angular


        self.volume = atoms.get_volume()
        self.pbc = atoms.get_pbc()
        self.Natoms = len(atoms)
        self.dim = 3

        # parameters for the binning:
        self.m1 = self.nsigma*self.sigma1/self.binwidth1  # number of neighbour bins included.
        self.smearing_norm1 = erf(1/np.sqrt(2) * self.m1 * self.binwidth1/self.sigma1)  # Integral of the included part of the gauss
        self.Nbins1 = int(np.ceil(self.Rc1/self.binwidth1))

        self.m2 = self.nsigma*self.sigma2/self.binwidth2  # number of neighbour bins included.
        self.smearing_norm2 = erf(1/np.sqrt(2) * self.m2 * self.binwidth2/self.sigma2)  # Integral of the included part of the gauss
        self.binwidth2 = np.pi/Nbins2

        # Prepare stuff to handle multiple species
        self.num = atoms.get_atomic_numbers()
        self.atomic_types = sorted(list(set(self.num)))
        self.atomic_count = {type:list(self.num).count(type) for type in self.atomic_types}
        self.Ntypes = len(self.atomic_types)

        type_converter = {}
        for i, type in enumerate(self.atomic_types):
            type_converter[type] = i

        # Bondtypes - 2body
        self.Nbondtypes_2body = 0
        self.bondtypes_2body = -np.ones((self.Ntypes, self.Ntypes)).astype(int)
        count = 0
        for tt1 in self.atomic_types:
            for tt2 in self.atomic_types:
                type1, type2 = tuple(sorted([tt1, tt2]))
                t1 = type_converter[type1]
                t2 = type_converter[type2]
                if type1 < type2:
                    if self.bondtypes_2body[t1,t2] == -1:
                        self.bondtypes_2body[t1,t2] = count
                        self.Nbondtypes_2body += 1
                        count += 1
                elif type1 == type2 and (self.atomic_count[type1] > 1 or sum(self.pbc) > 0):
                    if self.bondtypes_2body[t1,t2] == -1:
                        self.bondtypes_2body[t1,t2] = count
                        self.Nbondtypes_2body += 1
                        count += 1
        self.bondtypes_2body = self.bondtypes_2body.reshape(-1).tolist()


        # Bondtypes - 3body
        self.bondtypes_3body = -np.ones((self.Ntypes, self.Ntypes, self.Ntypes)).astype(int)
        bondtypes_3body_keys = []
        for type1 in self.atomic_types:
            for type2 in self.atomic_types:
                if type1 == type2 and not (self.atomic_count[type1] > 1 or sum(self.pbc) > 0):
                    continue
                for type3 in self.atomic_types:
                    if type1 == type3 and not (self.atomic_count[type1] > 1 or sum(self.pbc) > 0):
                        continue
                    key = tuple([type1] + sorted([type2, type3]))
                    if type2 < type3:
                        bondtypes_3body_keys.append(key)
                    elif type2 == type3:
                        if type2 == type1 and (self.atomic_count[type1] > 2 or sum(self.pbc) > 0):
                            bondtypes_3body_keys.append(key)
                        elif type2 != type1 and (self.atomic_count[type2] > 1 or sum(self.pbc) > 0):
                            bondtypes_3body_keys.append(key)
        self.Nbondtypes_3body = len(bondtypes_3body_keys)
        for i, key in enumerate(bondtypes_3body_keys):
            k1, k2, k3 = key
            self.bondtypes_3body[type_converter[k1],type_converter[k2],type_converter[k3]] = i
        self.bondtypes_3body = self.bondtypes_3body.reshape(-1).tolist()

        self.Nelements_2body = self.Nbondtypes_2body * self.Nbins1
        self.Nelements_3body = self.Nbondtypes_3body * self.Nbins2

        if use_angular:
            self.Nelements = self.Nelements_2body + self.Nelements_3body
        else:
            self.Nelements = self.Nelements_2body

        # Get relevant neighbour unit-cells
        self.pbc = atoms.get_pbc()
        self.cell = atoms.get_cell()
        self.neighbourcells_disp = self.__get_neighbour_cells_displacement(self.pbc, self.cell)

    def get_feature(self, atoms):
        """
        """
        cdef double Rc1 = self.Rc1
        cdef double Rc2 = self.Rc2
        cdef double binwidth1 = self.binwidth1
        cdef double binwidth2 = self.binwidth2
        cdef int Nbins1 = self.Nbins1
        cdef int Nbins2 = self.Nbins2
        cdef double sigma1 = self.sigma1
        cdef double sigma2 = self.sigma2
        cdef int nsigma = self.nsigma

        cdef double eta = self.eta
        cdef double gamma = self.gamma
        cdef use_angular = self.use_angular

        cdef double volume = self.volume
        cdef int dim = self.dim

        cdef double m1 = self.m1
        cdef double m2 = self.m2
        cdef double smearing_norm1 = self.smearing_norm1
        cdef double smearing_norm2 = self.smearing_norm2

        cdef int Nelements_2body = self.Nelements_2body
        cdef int Nelements_3body = self.Nelements_3body
        cdef int Nelements = self.Nelements

        # Memory allocation pool
        cdef Pool mem
        mem = Pool()

        cell = atoms.get_cell()
        cdef int Natoms = self.Natoms

        # Get positions and convert to Point-struct
        cdef list pos_np = atoms.get_positions().tolist()
        cdef Point *pos
        pos = <Point*>mem.alloc(Natoms, sizeof(Point))
        cdef int m
        for m in range(Natoms):
            pos[m].coord[0] = pos_np[m][0]
            pos[m].coord[1] = pos_np[m][1]
            pos[m].coord[2] = pos_np[m][2]

        # Get neighbourcells and convert to Point-struct
        cdef int Ncells = len(self.neighbourcells_disp)
        cdef list cell_displacements_old = self.neighbourcells_disp
        cdef Point *cell_displacements
        cell_displacements = <Point*>mem.alloc(Ncells, sizeof(Point))
        for m in range(Ncells):
            cell_displacements[m].coord[0] = cell_displacements_old[m][0]
            cell_displacements[m].coord[1] = cell_displacements_old[m][1]
            cell_displacements[m].coord[2] = cell_displacements_old[m][2]

        # Convert 2body bondtype list into c-array
        cdef int Ntypes = self.Ntypes
        cdef int Nbondtypes_2body = self.Nbondtypes_2body
        cdef list bondtypes_2body_old = self.bondtypes_2body
        cdef int *bondtypes_2body
        bondtypes_2body = <int*>mem.alloc(Ntypes*Ntypes, sizeof(int))
        for m in range(Ntypes*Ntypes):
            bondtypes_2body[m] = bondtypes_2body_old[m]

        # Get converted atom Ntypes
        cdef list num_converted_old = convert_atom_types(atoms.get_atomic_numbers())
        cdef int *num_converted
        num_converted = <int*>mem.alloc(Natoms, sizeof(int))
        for m in range(Natoms):
            num_converted[m] = num_converted_old[m]

        # RADIAL FEATURE

        # Initialize radial feature
        cdef double *feature1
        feature1 = <double*>mem.alloc(Nelements_2body, sizeof(double))

        cdef int num_pairs, center_bin, minbin_lim, maxbin_lim, newbin, bondtype_index, type1, type2
        cdef double Rij, normalization, binpos, c, erfarg_low, erfarg_up, value
        cdef int i, j, n
        for i in range(Natoms):
            for cell_index in range(Ncells):
                displacement = cell_displacements[cell_index]
                for j in range(Natoms):
                    Rij = euclidean(pos[i], add(pos[j], displacement))

                    # Stop if distance too long or atoms are the same one.
                    if Rij > Rc1+nsigma*sigma1 or Rij < 1e-6:
                        continue

                    # determine bondtype
                    if num_converted[i] <= num_converted[j]:
                        type1 = num_converted[i]
                        type2 = num_converted[j]
                    else:
                        type1 = num_converted[j]
                        type2 = num_converted[i]
                    bondtype_index = Nbins1*bondtypes_2body[Ntypes*type1+type2]

                    # Calculate normalization
                    num_pairs = Natoms*Natoms
                    normalization = 1./smearing_norm1
                    normalization /= 4*M_PI*Rij*Rij * binwidth1 * num_pairs/volume

                    # Identify what bin 'Rij' belongs to + it's position in this bin
                    center_bin = <int> floor(Rij/binwidth1)
                    binpos = Rij/binwidth1 - center_bin

                    # Lower and upper range of bins affected by the current atomic distance deltaR.
                    minbin_lim = <int> -ceil(m1 - binpos)
                    maxbin_lim = <int> ceil(m1 - (1-binpos))

                    for n in range(minbin_lim, maxbin_lim + 1):
                        newbin = center_bin + n
                        if newbin < 0 or newbin >= Nbins1:
                            continue

                        # Calculate gauss contribution to current bin
                        c = 1./sqrt(2)*binwidth1/sigma1
                        erfarg_low = max(-m1, n-binpos)
                        erfarg_up = min(m1, n+(1-binpos))
                        value = 0.5*erf(c*erfarg_up)-0.5*erf(c*erfarg_low)

                        # Apply normalization
                        value *= normalization

                        feature1[bondtype_index + newbin] += value
                        #feature1[newbin] += value

        # Convert radial feature to numpy array
        feature1_np = np.zeros(Nelements_2body)
        for m in range(Nelements_2body):
            feature1_np[m] = feature1[m]

        # Return feature if only radial part is desired
        if not use_angular:
            return feature1_np

        # ANGULAR FEATURE

        # Convert 3body bondtype list into c-array
        cdef int Nbondtypes_3body = self.Nbondtypes_3body
        cdef list bondtypes_3body_old = self.bondtypes_3body
        cdef int *bondtypes_3body
        bondtypes_3body = <int*>mem.alloc(Ntypes*Ntypes*Ntypes, sizeof(int))
        for m in range(Ntypes*Ntypes*Ntypes):
            bondtypes_3body[m] = bondtypes_3body_old[m]

        # Initialize angular feature
        cdef double *feature2
        feature2 = <double*>mem.alloc(Nelements_3body, sizeof(double))

        cdef Point RijVec, RikVec
        cdef double angle
        cdef int k, cond_ij, cond_ik, type3
        for i in range(Natoms):
            pos_i = pos[i]
            for cell_index1 in range(Ncells):
                displacement1 = cell_displacements[cell_index1]
                for j in range(Natoms):
                    pos_j = add(pos[j], displacement1)
                    Rij = euclidean(pos[i], pos_j)
                    if Rij > Rc2 or Rij < 1e-6:
                        continue
                    for cell_index2 in range(cell_index1, Ncells):
                        displacement2 = cell_displacements[cell_index2]
                        if cell_index1 == cell_index2:
                            k_start = j+1
                        else:
                            k_start = 0
                        for k in range(k_start, Natoms):
                            pos_k = add(pos[k], displacement2)
                            Rik = euclidean(pos_i, pos_k)
                            if Rik > Rc2 or Rik < 1e-6:
                                continue

                            # determine bondtype
                            type1 = num_converted[i]
                            if num_converted[j] <= num_converted[k]:
                                type2 = num_converted[j]
                                type3 = num_converted[k]
                            else:
                                type2 = num_converted[k]
                                type3 = num_converted[j]
                            bondtype_index = Nbins2*bondtypes_3body[Ntypes*Ntypes*type1 + Ntypes*type2 + type3]

                            # Calculate angle
                            RijVec = subtract(pos_j,pos_i)
                            RikVec = subtract(pos_k, pos_i)
                            angle = get_angle(RijVec, RikVec)

                            # Calculate normalization
                            num_pairs = Natoms*Natoms*Natoms
                            normalization = 1./smearing_norm2
                            normalization /= num_pairs/volume

                            # Identify what bin 'Rij' belongs to + it's position in this bin
                            center_bin = <int> floor(angle/binwidth1)
                            binpos = angle/binwidth2 - center_bin

                            # Lower and upper range of bins affected by the current atomic distance deltaR.
                            minbin_lim = <int> -ceil(m2 - binpos)
                            maxbin_lim = <int> ceil(m2 - (1-binpos))

                            for n in range(minbin_lim, maxbin_lim + 1):
                                newbin = center_bin + n

                                # Wrap current bin into correct bin-range
                                if newbin < 0:
                                    newbin = abs(newbin)
                                if newbin > Nbins2-1:
                                    newbin = 2*Nbins2 - newbin - 1

                                # Calculate gauss contribution to current bin
                                c = 1./sqrt(2)*binwidth2/sigma2
                                erfarg_low = max(-m2, n-binpos)
                                erfarg_up = min(m2, n+(1-binpos))
                                value = 0.5*erf(c*erfarg_up)-0.5*erf(c*erfarg_low)
                                value *= f_cutoff(Rij, gamma, Rc2) * f_cutoff(Rik, gamma, Rc2)
                                # Apply normalization
                                value *= normalization

                                feature2[bondtype_index + newbin] += value
                                #feature2[newbin] += value

        # Convert angular feature to numpy array
        feature2_np = np.zeros(Nelements_3body)
        for m in range(Nelements_3body):
            feature2_np[m] = eta * feature2[m]

        feature_np = np.zeros(Nelements)
        feature_np[:Nelements_2body] = feature1_np
        feature_np[Nelements_2body:] = feature2_np
        return feature_np

    def get_featureMat(self, atoms_list):
        featureMat = np.array([self.get_feature(atoms) for atoms in atoms_list])
        featureMat = np.array(featureMat)
        return featureMat

    def get_featureGradient(self, atoms):

        cdef double Rc1 = self.Rc1
        cdef double Rc2 = self.Rc2
        cdef double binwidth1 = self.binwidth1
        cdef double binwidth2 = self.binwidth2
        cdef int Nbins1 = self.Nbins1
        cdef int Nbins2 = self.Nbins2
        cdef double sigma1 = self.sigma1
        cdef double sigma2 = self.sigma2
        cdef int nsigma = self.nsigma

        cdef double eta = self.eta
        cdef double gamma = self.gamma
        cdef use_angular = self.use_angular

        cdef double volume = self.volume
        cdef int dim = self.dim

        cdef double m1 = self.m1
        cdef double m2 = self.m2
        cdef double smearing_norm1 = self.smearing_norm1
        cdef double smearing_norm2 = self.smearing_norm2

        cdef int Nelements_2body = self.Nelements_2body
        cdef int Nelements_3body = self.Nelements_3body
        cdef int Nelements = self.Nelements

        # Memory allocation pool
        cdef Pool mem
        mem = Pool()

        cell = atoms.get_cell()
        cdef int Natoms = self.Natoms

        # Get positions and convert to Point-struct
        cdef list pos_np = atoms.get_positions().tolist()
        cdef Point *pos
        pos = <Point*>mem.alloc(Natoms, sizeof(Point))
        cdef int m
        for m in range(Natoms):
            pos[m].coord[0] = pos_np[m][0]
            pos[m].coord[1] = pos_np[m][1]
            pos[m].coord[2] = pos_np[m][2]

        # Get neighbourcells and convert to Point-struct
        cdef int Ncells = len(self.neighbourcells_disp)
        cdef list cell_displacements_old = self.neighbourcells_disp
        cdef Point *cell_displacements
        cell_displacements = <Point*>mem.alloc(Ncells, sizeof(Point))
        for m in range(Ncells):
            cell_displacements[m].coord[0] = cell_displacements_old[m][0]
            cell_displacements[m].coord[1] = cell_displacements_old[m][1]
            cell_displacements[m].coord[2] = cell_displacements_old[m][2]

        # Convert bondtype list into c-array
        cdef int Ntypes = self.Ntypes
        cdef int Nbondtypes_2body = self.Nbondtypes_2body
        cdef list bondtypes_2body_old = self.bondtypes_2body
        cdef int *bondtypes_2body
        bondtypes_2body = <int*>mem.alloc(Ntypes*Ntypes, sizeof(int))
        for m in range(Ntypes*Ntypes):
            bondtypes_2body[m] = bondtypes_2body_old[m]

        # Get converted atom Ntypes
        cdef list num_converted_old = convert_atom_types(atoms.get_atomic_numbers())
        cdef int *num_converted
        num_converted = <int*>mem.alloc(Natoms, sizeof(int))
        for m in range(Natoms):
            num_converted[m] = num_converted_old[m]

        # RADIAL FEATURE GRADIENT

        # Initialize radial feature-gradient
        cdef double *feature_grad1
        feature_grad1 = <double*>mem.alloc(Nelements_2body * Natoms * dim, sizeof(double))

        cdef Point RijVec
        cdef int num_pairs, center_bin, minbin_lim, maxbin_lim, newbin, bondtype_index, type1, type2
        cdef double Rij, normalization, binpos, c, arg_low, arg_up, value1, value2, value
        cdef int i, j, n
        for i in range(Natoms):
            pos_i = pos[i]
            for cell_index in range(Ncells):
                displacement = cell_displacements[cell_index]
                for j in range(Natoms):
                    pos_j = add(pos[j], displacement)
                    Rij = euclidean(pos_i, pos_j)

                    # Stop if distance too long or atoms are the same one.
                    if Rij > Rc1+nsigma*sigma1 or Rij < 1e-6:
                        continue
                    RijVec = subtract(pos_j,pos_i)

                    # determine bondtype
                    if num_converted[i] <= num_converted[j]:
                        type1 = num_converted[i]
                        type2 = num_converted[j]
                    else:
                        type1 = num_converted[j]
                        type2 = num_converted[i]
                    bondtype_index = Nbins1*bondtypes_2body[Ntypes*type1+type2]

                    # Calculate normalization
                    num_pairs = Natoms*Natoms
                    normalization = 1./smearing_norm1
                    normalization /= 4*M_PI*Rij*Rij * binwidth1 * num_pairs/volume

                    # Identify what bin 'Rij' belongs to + it's position in this bin
                    center_bin = <int> floor(Rij/binwidth1)
                    binpos = Rij/binwidth1 - center_bin

                    # Lower and upper range of bins affected by the current atomic distance deltaR.
                    minbin_lim = <int> -ceil(m1 - binpos)
                    maxbin_lim = <int> ceil(m1 - (1-binpos))

                    for n in range(minbin_lim, maxbin_lim + 1):
                        newbin = center_bin + n
                        if newbin < 0 or newbin >= Nbins1:
                            continue

                        # Calculate gauss contribution to current bin
                        c = 1./sqrt(2)*binwidth1/sigma1
                        arg_low = max(-m1, n-binpos)
                        arg_up = min(m1, n+(1-binpos))
                        value1 = -1./Rij * (erf(c*arg_up) - erf(c*arg_low))
                        value2 = -1./(sigma1*sqrt(2*M_PI)) * (exp(-pow(c*arg_up,2)) - exp(-pow(c*arg_low,2)))
                        value = value1 + value2

                        # Apply normalization
                        value *= normalization

                        # Add to the the gradient matrix
                        for m in range(3):
                            feature_grad1[(bondtype_index + newbin) * Natoms*dim + dim*i+m] += -value/Rij * RijVec.coord[m]
                            feature_grad1[(bondtype_index + newbin) * Natoms*dim + dim*j+m] += value/Rij * RijVec.coord[m]


        # Convert radial feature to numpy array
        cdef int grad_index
        feature_grad1_np = np.zeros((Natoms*dim, Nelements_2body))
        for m in range(Nelements_2body):
            for grad_index in range(Natoms*dim):
                feature_grad1_np[grad_index][m] = feature_grad1[m * Natoms*dim + grad_index]

        # Return feature if only radial part is desired
        if not use_angular:
            return feature_grad1_np

        # ANGULAR FEATURE-GRADIENT

        # Convert 3body bondtype list into c-array
        cdef int Nbondtypes_3body = self.Nbondtypes_3body
        cdef list bondtypes_3body_old = self.bondtypes_3body
        cdef int *bondtypes_3body
        bondtypes_3body = <int*>mem.alloc(Ntypes*Ntypes*Ntypes, sizeof(int))
        for m in range(Ntypes*Ntypes*Ntypes):
            bondtypes_3body[m] = bondtypes_3body_old[m]

        # Initialize angular feature-gradient
        cdef double *feature_grad2
        feature_grad2 = <double*>mem.alloc(Nelements_3body * Natoms * dim, sizeof(double))

        cdef Point RikVec, angle_grad_i, angle_grad_j, angle_grad_k
        cdef double angle, cos_angle, a
        cdef int k, cond_ij, cond_ik, bin_index, type3
        for i in range(Natoms):
            pos_i = pos[i]
            for cell_index1 in range(Ncells):
                displacement1 = cell_displacements[cell_index1]
                for j in range(Natoms):
                    pos_j = add(pos[j], displacement1)
                    Rij = euclidean(pos[i], pos_j)
                    if Rij > Rc2 or Rij < 1e-6:
                        continue
                    for cell_index2 in range(cell_index1, Ncells):
                        displacement2 = cell_displacements[cell_index2]
                        if cell_index1 == cell_index2:
                            k_start = j+1
                        else:
                            k_start = 0
                        for k in range(k_start, Natoms):
                            pos_k = add(pos[k], displacement2)
                            Rik = euclidean(pos_i, pos_k)
                            if Rik > Rc2 or Rik < 1e-6:
                                continue

                            # determine bondtype
                            type1 = num_converted[i]
                            if num_converted[j] <= num_converted[k]:
                                type2 = num_converted[j]
                                type3 = num_converted[k]
                            else:
                                type2 = num_converted[k]
                                type3 = num_converted[j]
                            bondtype_index = Nbins2*bondtypes_3body[Ntypes*Ntypes*type1 + Ntypes*type2 + type3]

                            # Calculate angle
                            RijVec = subtract(pos_j,pos_i)
                            RikVec = subtract(pos_k, pos_i)


                            angle = get_angle(RijVec, RikVec)
                            cos_angle = cos(angle)

                            for m in range(3):
                                if not (angle == 0 or angle == M_PI):
                                    a = -1/sqrt(1 - cos_angle*cos_angle)
                                    angle_grad_j.coord[m] = a * (RikVec.coord[m]/(Rij*Rik) - cos_angle*RijVec.coord[m]/(Rij*Rij))
                                    angle_grad_k.coord[m] = a * (RijVec.coord[m]/(Rij*Rik) - cos_angle*RikVec.coord[m]/(Rik*Rik))
                                    angle_grad_i.coord[m] = -(angle_grad_j.coord[m] + angle_grad_k.coord[m])
                                else:
                                    angle_grad_j.coord[m] = 0
                                    angle_grad_k.coord[m] = 0
                                    angle_grad_i.coord[m] = 0

                            fc_ij = f_cutoff(Rij, gamma, Rc2)
                            fc_ik = f_cutoff(Rik, gamma, Rc2)
                            fc_grad_ij = f_cutoff_grad(Rij, gamma, Rc2)
                            fc_grad_ik = f_cutoff_grad(Rik, gamma, Rc2)

                            # Calculate normalization
                            num_pairs = Natoms*Natoms*Natoms
                            normalization = 1./smearing_norm2
                            normalization /= num_pairs/volume

                            # Identify what bin 'Rij' belongs to + it's position in this bin
                            center_bin = <int> floor(angle/binwidth1)
                            binpos = angle/binwidth2 - center_bin

                            # Lower and upper range of bins affected by the current atomic distance deltaR.
                            minbin_lim = <int> -ceil(m2 - binpos)
                            maxbin_lim = <int> ceil(m2 - (1-binpos))

                            for n in range(minbin_lim, maxbin_lim + 1):
                                newbin = center_bin + n

                                # Wrap current bin into correct bin-range
                                if newbin < 0:
                                    newbin = abs(newbin)
                                if newbin > Nbins2-1:
                                    newbin = 2*Nbins2 - newbin - 1

                                # Calculate gauss contribution to current bin
                                c = 1./sqrt(2)*binwidth2/sigma2
                                arg_low = max(-m2, n-binpos)
                                arg_up = min(m2, n+(1-binpos))
                                value1 = 0.5*erf(c*arg_up)-0.5*erf(c*arg_low)
                                value2 = -1./(sigma2*sqrt(2*M_PI)) * (exp(-pow(c*arg_up, 2)) - exp(-pow(c*arg_low, 2)))

                                # Apply normalization
                                value1 *= normalization
                                value2 *= normalization

                                bin_index = (bondtype_index + newbin) * Natoms*dim
                                for m in range(3):
                                    feature_grad2[bin_index + dim*i+m] += -value1 * fc_ik*fc_grad_ij * RijVec.coord[m]/Rij
                                    feature_grad2[bin_index + dim*j+m] += value1 * fc_ik*fc_grad_ij * RijVec.coord[m]/Rij

                                    feature_grad2[bin_index + dim*i+m] += -value1 * fc_ij*fc_grad_ik * RikVec.coord[m]/Rik
                                    feature_grad2[bin_index + dim*k+m] += value1 * fc_ij*fc_grad_ik * RikVec.coord[m]/Rik

                                    feature_grad2[bin_index + dim*i+m] += value2 * fc_ij * fc_ik * angle_grad_i.coord[m]
                                    feature_grad2[bin_index + dim*j+m] += value2 * fc_ij * fc_ik * angle_grad_j.coord[m]
                                    feature_grad2[bin_index + dim*k+m] += value2 * fc_ij * fc_ik * angle_grad_k.coord[m]


        feature_grad2_np = np.zeros((Natoms*dim, Nelements_3body))
        for m in range(Nelements_3body):
            for grad_index in range(Natoms*dim):
                feature_grad2_np[grad_index][m] = eta * feature_grad2[m * Natoms*dim + grad_index]

        feature_grad_np = np.zeros((Natoms*dim, Nelements))
        feature_grad_np[:, :Nelements_2body] = feature_grad1_np
        feature_grad_np[:, Nelements_2body:] = feature_grad2_np
        return feature_grad_np

    def get_all_featureGradients(self, atoms_list):
        feature_grads = np.array([self.get_featureGradient(atoms) for atoms in atoms_list])
        feature_grads = np.array(feature_grads)
        return feature_grads

    def __get_neighbour_cells_displacement(self, pbc, cell):
        # Calculate neighbour cells
        Rc_max = max(self.Rc1+self.sigma1*self.nsigma, self.Rc2)  # relevant cutoff
        cell_vec_norms = np.linalg.norm(cell, axis=0)
        neighbours = []
        for i in range(3):
            if pbc[i]:
                ncellmax = int(np.ceil(abs(Rc_max/cell_vec_norms[i]))) + 1  # +1 because atoms can be outside unitcell.
                neighbours.append(range(-ncellmax,ncellmax+1))
            else:
                neighbours.append([0])

        neighbourcells_disp = []
        for x,y,z in product(*neighbours):
            xyz = (x,y,z)
            displacement = np.dot(xyz, cell)
            neighbourcells_disp.append(list(displacement))

        return neighbourcells_disp



from pathlib import Path
from typing import Callable, Dict, List, Tuple

import numpy as np
import pytest
from ase.symbols import Symbols

from agox.candidates import Candidate
from agox.databases import Database
from agox.environments import Environment


@pytest.fixture(scope='module')
def database_path(tmpdir_factory: Callable, environment_and_dataset: Tuple[Environment, List[Candidate]]) -> str:
    environment, dataset = environment_and_dataset
    total_symbols = Symbols(environment.get_all_numbers())

    # Make and move to temp dir:
    d = tmpdir_factory.mktemp("database_test")
    return d / f"db_{total_symbols}.db"

@pytest.fixture(scope='module')
def meta_dict() -> Dict:
    return {
        "str": "str",
        "int": 0,
        "float": 0.1,
        "1darr": np.array([0, 1, 2]),
        "2darr": np.random.rand(3, 3)
    }

@pytest.fixture(scope='module')
def dataset(environment_and_dataset: Tuple[Environment, List[Candidate]], meta_dict: Dict) -> list:
    environment, dataset = environment_and_dataset

    meta_data_names = list(meta_dict.keys())
    meta_data_examples = list(meta_dict.values())

    with_meta_data = []
    for candidate in dataset:
        candidate = candidate.copy()
        for meta_data_name, meta_data in zip(meta_data_names, meta_data_examples):
            candidate.add_meta_information(meta_data_name, meta_data)
        with_meta_data.append(candidate)
    return dataset

@pytest.fixture(scope='module')
def database_created(database_path: Path, dataset: list, meta_dict: dict) -> Database:
    # Make and move to temp dir:
    database_created = Database(database_path)
    meta_data_names = list(meta_dict.keys())
    meta_data_examples = list(meta_dict.values())

    for candidate in dataset:
        for meta_data_name, meta_data in zip(meta_data_names, meta_data_examples):
            candidate.add_meta_information(meta_data_name, meta_data)
        database_created.store_candidate(candidate)

    return database_created

@pytest.fixture(scope='module')
def database_loaded(database_path: Path) -> Database:
    database_loaded = Database(database_path)
    database_loaded.restore_to_memory()
    return database_loaded


from pathlib import Path

from click.testing import CliRunner

from agox.cli.main import main
from agox.test.test_utils import TemporaryFolder


def test_plot_cli_help() -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["plot", "--help"])
    assert result.exit_code == 0


def test_plot_cli_database(database_file_path: Path) -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["plot", str(database_file_path.resolve()), "--backend", "agg"])
    assert result.exit_code == 0


def test_plot_cli_trajectory(trajectory_file_path: Path) -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["plot", str(trajectory_file_path.resolve()), "--backend", "agg"])
    assert result.exit_code == 0


def test_plot_cli_save(tmpdir, database_file_path: Path) -> None:
    with TemporaryFolder(tmpdir):
        output = "fig.png"
        runner = CliRunner()
        result = runner.invoke(
            main, ["plot", str(database_file_path.resolve()), "--backend", "agg", "--output", output]
        )

        assert Path(output).exists()

    assert result.exit_code == 0




import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators.MD import MDgenerator
from agox.samplers import MetropolisSampler

# Manually set seed and database-index
seed = 42
database_index = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# database_index = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au2Ni2",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

sampler = MetropolisSampler(temperature=0.25, order=3)

generator = MDgenerator(
    **environment.get_confinement(),
    calculator=calc,
    temperature_program=[(50, 5)],  # 50 K for 5 steps
    constraints=environment.get_constraints(),
    environment=environment,
    sampler=sampler,
    order=1,
)

# Do single-point relaxation after MD
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"steps": 0},
    store_trajectory=False,
    order=2,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(generator, database, sampler, evaluator, seed=seed)

agox.run(N_iterations=10)



import matplotlib

matplotlib.use("Agg")


import numpy as np
from ase import Atoms

from agox import AGOX
from agox.collectors import StandardCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator, RattleGenerator
from agox.models.descriptors.simple_fingerprint import SimpleFingerprint
from agox.samplers import DistanceComparator, GeneticSampler

# Manually set seed and database-index
seed = 42
database_index = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# database_index = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au8Ni8",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

population_size = 10
descriptor = SimpleFingerprint(environment=environment)
comparator = DistanceComparator(descriptor, threshold=0.5)
sampler = GeneticSampler(population_size=population_size, comparator=comparator, order=4, database=database)

rattle_generator = RattleGenerator(**environment.get_confinement())
random_generator = RandomGenerator(**environment.get_confinement())
generators = [random_generator, rattle_generator]
num_candidates = {
    0: [population_size, 0],
    5: [2, population_size - 2],
    10: [0, population_size],
}

collector = StandardCollector(
    generators=generators,
    sampler=sampler,
    environment=environment,
    num_candidates=num_candidates,
    order=1,
)

# Very few steps used here - Set to higher number (>100) for a real run.
# The number fo evaluate is equal to to population size - which is also
# the total number of structures generated pr. iteration.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    store_trajectory=False,
    order=2,
    constraints=environment.get_constraints(),
    number_to_evaluate=population_size,
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(collector, database, evaluator, seed=seed)

agox.run(N_iterations=5)



import numpy as np
import pytest
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator

from agox.models import CompositionModel


@pytest.fixture
def dataset() -> list[Atoms]:
    data = []
    test_atom_energies = np.zeros(100)
    test_atom_energies[8] = 10
    test_atom_energies[6] = 5

    for _ in range(10):
        numbers = np.random.choice([6, 8], 10, replace=True)
        energy = np.sum(test_atom_energies[numbers])
        atoms = Atoms(numbers)
        atoms.calc = SinglePointCalculator(atoms, energy=energy, forces=np.zeros((len(atoms), 3)))
        data.append(atoms)

    return data

@pytest.fixture
def model(dataset: list[Atoms]) -> CompositionModel:
    model = CompositionModel()
    model.train(dataset)
    return model

def test_train(model: CompositionModel) -> None:
    assert model.atom_energies[8] == pytest.approx(10, abs=1e-3)
    assert model.atom_energies[6] == pytest.approx(5, abs=1e-3)
    mask = np.ones(100, dtype=bool)
    mask[6] = False
    mask[8] = False
    assert np.all(model.atom_energies[mask] == 0)

def test_predict_energy(model: CompositionModel) -> None:
    atoms = Atoms([6, 8, 6, 8])
    assert model.predict_energy(atoms) == pytest.approx(30, abs=1e-3)

def test_predict_forces(model: CompositionModel) -> None:
    atoms = Atoms([6, 8, 6, 8])
    assert np.all(model.predict_forces(atoms) == 0)


import numpy as np
from ase import Atoms

from agox.candidates import CandidateBaseClass
from agox.environments.environment import Environment
from agox.generators import SymmetryGenerator, SymmetryPermutationGenerator, SymmetryRattleGenerator
from agox.samplers.fixed_sampler import FixedSampler


def environment_factory():
    template = Atoms("", cell=np.eye(3) * 22, pbc=False)
    confinement_corner = np.array([5.5, 5.5, 5.5])
    confinement_cell = np.eye(3) * 11
    environment = Environment(
        template=template, symbols="B12N12", confinement_corner=confinement_corner, confinement_cell=confinement_cell
    )
    return environment


def generator_factory(environment):
    generator = SymmetryGenerator(
        sym_type="cluster", force_group="D6", may_nucleate_at_several_places=True, **environment.get_confinement()
    )
    return generator


def rattle_generator_factory(environment):
    rattle_generator = SymmetryRattleGenerator(**environment.get_confinement(), let_lose_sym=1.0)
    return rattle_generator


def permutation_generator_factory(environment):
    permutation_generator = SymmetryPermutationGenerator(**environment.get_confinement())
    return permutation_generator


def test_symmetry_generator():
    np.random.seed(1)
    environment = environment_factory()
    generator = generator_factory(environment)
    candidate = generator(None, environment)[0]
    assert isinstance(candidate, CandidateBaseClass)


def test_symmetry_rattle_generator():
    np.random.seed(1)
    environment = environment_factory()
    generator = generator_factory(environment)
    rattle_generator = rattle_generator_factory(environment)
    symmetry_candidate = generator(None, environment)[0]
    sampler = FixedSampler(symmetry_candidate)
    candidate = rattle_generator(sampler, environment)[0]
    assert isinstance(candidate, CandidateBaseClass)


def test_symmetry_permutation_generator():
    np.random.seed(1)
    environment = environment_factory()
    generator = generator_factory(environment)
    permutation_generator = permutation_generator_factory(environment)
    symmetry_candidate = generator(None, environment)[0]
    sampler = FixedSampler(symmetry_candidate)
    candidate = permutation_generator(sampler, environment)[0]
    assert isinstance(candidate, CandidateBaseClass)




from agox.test.run_tests.run_utils import agox_test_run


# @pytest.mark.skip(reason='Not reproducible on CI at the moment.')
def test_rss_cluster_symmetry(tmp_path, cmd_options):
    mode = "rss_cluster_symmetry"  # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options, full_compare=False)



import os

import numpy as np
import pytest

from agox.models.datasets import datasets


def make_local_gpr():
    from agox.models.descriptors import SOAP
    from agox.models.GPR import SparseGPR
    from agox.models.GPR.kernels import RBF
    from agox.models.GPR.kernels import Constant as C

    kernel = C(1, (1, 100)) * RBF(10, (10, 30))
    descriptor = SOAP.from_species(species=["Ag", "O"])
    model = SparseGPR(
        kernel=kernel,
        descriptor=descriptor,
        unc_jitter=1e-3,
        train_uncertainty=True,
    )
    return model


def make_global_gpr():
    from agox.models.descriptors import Fingerprint
    from agox.models.GPR import GPR
    from agox.models.GPR.kernels import RBF, Noise
    from agox.models.GPR.kernels import Constant as C
    from agox.models.GPR.priors import Repulsive

    descriptor = Fingerprint.from_atoms(datasets["Ag5O3"][0])
    kernel = C() * RBF() + Noise(0.01)
    model = GPR(descriptor, kernel, prior=Repulsive(), n_optimize=0)
    return model


@pytest.mark.ray
@pytest.mark.parametrize("model_maker", [make_global_gpr, make_local_gpr])
def test_model(model_maker, cmd_options, tmpdir):
    tolerances = cmd_options["tolerance"]

    data = datasets["Ag5O3"]
    training_data = data[:-1]
    test_data = data[-1]

    model = model_maker()
    model.train(training_data)

    E = model.predict_energy(test_data)
    F = model.predict_forces(test_data)
    S = model.predict_uncertainty(test_data)

    tmpfile = str(tmpdir.join("test_model.h5"))
    model.save(tmpfile)

    new_model = model_maker()
    new_model.load(tmpfile)
    os.remove(tmpfile)

    E_new = new_model.predict_energy(test_data)
    F_new = new_model.predict_forces(test_data)
    S_new = new_model.predict_uncertainty(test_data)

    np.testing.assert_allclose(E, E_new, **tolerances)
    np.testing.assert_allclose(F, F_new, **tolerances)
    np.testing.assert_allclose(S, S_new, **tolerances)



import numpy as np
from ase import Atoms
from chgnet.model.dynamics import CHGNetCalculator

from agox import AGOX
from agox.acquisitors import ReplicaExchangeAcquisitor
from agox.collectors import ReplicaExchangeCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import SinglePointEvaluator
from agox.helpers.gpaw_subprocess import SubprocessGPAW
from agox.models import CalculatorModel
from agox.models.descriptors import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.postprocessors import ParallelRelaxPostprocess
from agox.samplers import ReplicaExchangeSampler
from agox.utils.sparsifiers.CUR import CUR

##############################################################################
# General settings:
##############################################################################

seed = 42            # Random seed for reproducibility - should be set to a unique value for each run. 
database_index = 0   # Index for database file - should be unique for each run to avoid overwriting previous results.
n_iterations = 1000  # Number of iterations to run the AGOX loop.

##############################################################################
# Replica exchange settings:
##############################################################################

temperature_range = [0.02, 2.0]  # Temperature range for the walkers in eV.
sample_size = 10  # Number of walkers in the replica exchange simulation.
swap_interval = 10  # How often to attempt a swap between walkers.
rattle_amplitudes = np.linspace(0.1, 5, sample_size)  # Rattle amplitudes for the walkers.
dft_interval = 100  # Iterations between DFT evaluations.
n_dft = 5  # Number of structures to evaluate with DFT when selected.

##############################################################################
# Calculator
##############################################################################

calculator = SubprocessGPAW(
    mode={"name": "pw", "ecut": 500},
    xc="PBE",
    maxiter=200,
    kpts=(1, 1, 1),
    convergence={"energy": 0.0001, "density": 1.0e-4, "eigenstates": 1.0e-4, "bands": "occupied"},
    occupations={"name": "fermi-dirac", "width": 0.1},
    nbands="110%",
)

##############################################################################
# System & general settings:
##############################################################################

symbols = "Ag16S8"
cell_size = 20.0
confinement_size = 15.0
template = Atoms(cell=np.eye(3) * cell_size)
confinement_cell = np.eye(3) * confinement_size
confinement_corner = np.ones(3) * (cell_size - confinement_size) / 2
environment = Environment(
    template=template,
    symbols=symbols,
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
    box_constraint_pbc=[False, False, False],
)

constraints = environment.get_constraints()

# Database
db_path = "db{}.db".format(database_index)
database = Database(filename=db_path, order=6)

##############################################################################
# Machine learning surrogate model:
##############################################################################

# Prior - Pretrained CHGNet model.
prior = CalculatorModel(calculator=CHGNetCalculator(on_isolated_atoms="ignore"))

# Descriptor
symbols = environment.get_all_species()
descriptor = SOAP.from_species(symbols, r_cut=7.0)

# GPR Kernel
kernel = C(1) * RBF(40)
noise = 0.01

sparsifier = CUR(1000)

model = SparseGPR(
    database=database,
    kernel=kernel,
    descriptor=descriptor,
    prior=prior,
    n_optimize=0,
    sparsifier=sparsifier,
    noise=noise,
    centralize=True,
)

##############################################################################
# Search Settings:
##############################################################################

sampler = ReplicaExchangeSampler(
    model=model,
    t_min=temperature_range[0],
    t_max=temperature_range[1],
    swap_interval=swap_interval,
    sample_size=sample_size,
    order=3,
    verbosity=2,
)

collector = ReplicaExchangeCollector.from_sampler(
    sampler,
    environment,
    rattle_amplitudes,
)

# If any generators will have dynamic parameters i.e. rattle amplitude, define them here
for i in range(sample_size - 1):
    collector.add_generator_update(i, "rattle_amplitude", min_val=0.01, max_val=5)

# Define an acquisitor which manages what and when structures are passed to the evaluator
acquisitor = ReplicaExchangeAcquisitor.with_interval(dft_interval=dft_interval)

relaxer = ParallelRelaxPostprocess(
    model=model,
    constraints=constraints,
    optimizer_run_kwargs={"steps": 100, "fmax": 0.025},  # Choose this appropriately.
    start_relax=0,
    order=2,
)

evaluator = SinglePointEvaluator(calculator, number_to_evaluate=n_dft)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(collector, relaxer, acquisitor, evaluator, sampler, database, seed=seed)

agox.run(N_iterations=n_iterations)




import pytest



@pytest.mark.ray
def test_local(tmp_path):
    pass


# @pytest.mark.ray
# def test_local_forces(tmp_path):
#     with TemporaryFolder(tmp_path):
#         import agox.test.model_tests.sgpr_api_forces

# @pytest.mark.ray
# def test_global(tmp_path):
#     with TemporaryFolder(tmp_path):
#         import agox.test.model_tests.gpr_api

# @pytest.mark.ray
# def test_descriptors(tmp_path):
#     with TemporaryFolder(tmp_path):
#         import agox.test.model_tests.descriptors_api



from agox.test.run_tests.run_utils import agox_test_run
import pytest 

@pytest.mark.skip(reason="Somehow doesn't work on CI")
def test_md_bh(tmp_path, cmd_options):
    mode = "md_bh"  # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options)



from agox.test.run_tests.run_utils import agox_test_run


def test_block_bh(tmp_path, cmd_options):
    mode = "block_bh"  # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options)



import numpy as np
import pytest

from agox.models.descriptors import Fingerprint
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.test.model_tests.model_utils import model_tester
from agox.test.test_utils import (
    check_file_is_deleted,
    get_name,
    get_test_data,
    get_test_environment,
    label_dict_list,
    load_expected_data,
    save_expected_data,
    test_data_dicts,
)

# Weird kernel specification is because of accurcay in Sklearn that does such a conversion.
kernel = C(1.0) * RBF(length_scale=np.exp(np.log(20.0)))
model_class = SparseGPR
model_maker = SparseGPR
model_base_args = []
model_base_kwargs = {}

model_update_kwargs = [
    {
        "kernel": kernel,
        "noise": 0.2,
        "database": None,
        "use_ray": False,
        "jitter": 5e-3,
    },
]

model_update_kwargs = label_dict_list(model_update_kwargs)


@pytest.fixture(params=model_update_kwargs)
def update_kwargs(request):
    return request.param


@pytest.mark.ray
@pytest.mark.parametrize("test_data_dict", test_data_dicts)
def test_model(test_data_dict, update_kwargs, cmd_options):
    # pytest.skip('This test is not yet working, due to tolerances on different machines making it unreliable.')

    create_mode = cmd_options["create_mode"]
    test_mode = not create_mode
    tolerance = cmd_options["tolerance"]

    update_kwargs, parameter_index = update_kwargs

    path = test_data_dict["path"]
    remove = test_data_dict["remove"]
    dataset_name = test_data_dict["name"]
    parameter_index = 0

    # Gather environment and data:
    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)

    # Update the kwargs for the model.
    descriptor = Fingerprint(
        environment=environment,
    )
    update_kwargs["descriptor"] = descriptor
    # Slightly complicated way of building input args & kwargs:
    if "environment" in update_kwargs.keys():
        update_kwargs["environment"] = environment

    model_kwargs = model_base_kwargs.copy()
    model_kwargs.update(update_kwargs)

    # Where to load/save data.
    subfolder = "model_tests/"
    module_name = "globalSparseGPR"
    name = get_name(module_name, subfolder, dataset_name, parameter_index)
    if test_mode:
        expected_data = load_expected_data(name)
    else:
        expected_data = None

    output = model_tester(
        model_maker,
        model_base_args,
        model_kwargs,
        data,
        test_mode=test_mode,
        expected_data=expected_data,
        tolerance=tolerance,
    )

    if not test_mode:
        check_file_is_deleted(name)
        save_expected_data(name, output)



import pytest

from agox.models.descriptors import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.test.model_tests.model_utils import model_tester
from agox.test.test_utils import (
    check_file_is_deleted,
    get_name,
    get_test_data,
    get_test_environment,
    label_dict_list,
    load_expected_data,
    save_expected_data,
    test_data_dicts,
)

kernel = C() * RBF() + Noise(0.01)
model_class = GPR
model_maker = GPR
model_base_args = []
model_base_kwargs = {"use_ray": False}

model_update_kwargs = [{"kernel": kernel}]

model_update_kwargs = label_dict_list(model_update_kwargs)


@pytest.fixture(params=model_update_kwargs)
def update_kwargs(request):
    return request.param


@pytest.mark.ray
@pytest.mark.parametrize("test_data_dict", test_data_dicts)
def test_model(test_data_dict, update_kwargs, cmd_options):
    create_mode = cmd_options["create_mode"]
    test_mode = not create_mode
    tolerance = cmd_options["tolerance"]

    update_kwargs, parameter_index = update_kwargs

    path = test_data_dict["path"]
    remove = test_data_dict["remove"]
    dataset_name = test_data_dict["name"]
    parameter_index = 0

    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)

    # Update the kwargs for the model.
    descriptor = Fingerprint(environment=environment)
    update_kwargs["descriptor"] = descriptor

    model_kwargs = model_base_kwargs.copy()
    model_kwargs.update(update_kwargs)

    # Where to load/save data.
    subfolder = "model_tests/"
    module_name = "GlobalGPR"
    name = get_name(module_name, subfolder, dataset_name, parameter_index)
    if test_mode:
        expected_data = load_expected_data(name)
    else:
        expected_data = None

    output = model_tester(
        model_maker,
        model_base_args,
        model_kwargs,
        data,
        test_mode=test_mode,
        expected_data=expected_data,
        tolerance=tolerance,
    )

    if not test_mode:
        check_file_is_deleted(name)
        save_expected_data(name, output)



import pytest

from agox.generators import RandomGenerator
from agox.test.generator_tests.generator_utils import generator_testing
from agox.test.test_utils import test_data_dicts

seed = 1
generator_args = []
generator_base_kwargs = {"c1": 0.75, "c2": 1.25}
generator_class = RandomGenerator

list_of_other_kwargs = [
    {},  # Tests that defaults havent changed.
    {"contiguous": True},
    {"contiguous": False},
]

for index, dictionary in enumerate(list_of_other_kwargs):
    list_of_other_kwargs[index] = (dictionary, index)


@pytest.fixture(params=list_of_other_kwargs)
def other_kwargs(request):
    return request.param


@pytest.mark.parametrize("test_data_dict", test_data_dicts)
def test_generator(test_data_dict, other_kwargs, cmd_options):
    parameter_index = other_kwargs[1]
    other_kwargs = other_kwargs[0]

    create_mode = cmd_options["create_mode"]
    test_mode = not create_mode
    tolerance = cmd_options["tolerance"]

    generator_testing(
        generator_class,
        test_data_dict,
        generator_args,
        generator_base_kwargs,
        other_kwargs,
        parameter_index,
        tolerance,
        seed=seed,
        test_mode=test_mode,
    )




import matplotlib

matplotlib.use("Agg")


import numpy as np
from ase import Atoms

from agox import AGOX
from agox.databases import ConcurrentDatabase
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RattleGenerator
from agox.samplers import ConcurrentTemperingSampler

# Manually set seed and run_idx
seed = 42
run_idx = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# run_idx = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au12Ni2",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
total_workers = 1  # This should be higher for a real run.
sync_frequency = 10
database_index = (run_idx - 1) // total_workers + 1
worker_index = (run_idx - 1) % total_workers
db_path = "db{}.db".format(database_index)
database = ConcurrentDatabase(
    filename=db_path,
    store_meta_information=True,
    worker_number=worker_index,
    total_workers=total_workers,
    sync_frequency=sync_frequency,
    order=4,
    sleep_timing=0.1,
)

# ##############################################################################
# # Search Settings:
# ##############################################################################

temperatures = [0.1 * 1.5**power for power in range(total_workers)]
sampler = ConcurrentTemperingSampler(temperatures=temperatures, order=3, database=database, swap_order=5)

rattle_generator = RattleGenerator(**environment.get_confinement(), environment=environment, sampler=sampler, order=1)

# Very few steps used here - Set to higher number (>100) for a real run.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    store_trajectory=False,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    order=2,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(rattle_generator, database, evaluator, sampler, seed=seed)

agox.run(N_iterations=20)



from agox.models.descriptors import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.utils.sparsifiers import CUR

kernel = C(1, (1, 100)) * RBF(20, (10, 30))
descriptor = SOAP.from_species(
    species=["Ag", "O"],
    r_cut=5,
    nmax=3,
    lmax=2,
    sigma=1,
    weight=True,
    periodic=False,
)
model = SparseGPR(
    kernel=kernel,
    descriptor=descriptor,
    sparsifier=CUR(1000),  # sparsification using CUR algorithm with 1000 sparse points
    noise=0.01,  # eV/atom
)

# load the model from a file
model.load("sgpr.h5")



import pytest
from ase.io import read

from agox.models.descriptors.spectral_graph_descriptor import SpectralGraphDescriptor
from agox.test.test_utils import test_data_dicts
from agox.utils.filters.feature_distance import FeatureDistanceFilter


@pytest.mark.ray
@pytest.mark.parametrize("test_data_dict", test_data_dicts)
def test_feature_distance_filter(test_data_dict):
    path = test_data_dict["path"]

    # Load structures from directory
    all_structures = read(path, ":")

    # Setup a descriptor
    descriptor = SpectralGraphDescriptor.from_atoms(all_structures[0])

    # Setup a filter, structures closer than threshold are considered duplicates
    threshold = 1e-8
    feature_filter = FeatureDistanceFilter(descriptor, threshold)

    # Filter the structures
    filtered_structures, filtered_indices = feature_filter.filter(all_structures)

    print("Before filtering: ", len(all_structures))
    print("After filtering: ", len(filtered_structures))



import os
from copy import deepcopy
from importlib.resources import files
from typing import List, Tuple

import numpy as np
import pytest
from ase.io import read

from agox.candidates import StandardCandidate
from agox.environments.environment import Environment


def pytest_addoption(parser):
    parser.addoption("--rtol", type=float, default=1e-05)
    parser.addoption("--atol", type=float, default=1e-08)
    parser.addoption("--create_mode", default=False, action="store_true")

class EnvironmentFactory:

    def __init__(self, environment: Environment) -> None:
        self.environment = environment
    
    def __call__(self) -> Environment:
        return self.environment
    
class DatasetFactory:
    
    def __init__(self, candidates: List[StandardCandidate]) -> None:
        self.candidates = candidates
    
    def __call__(self) -> List[StandardCandidate]:
        return deepcopy(self.candidates)



@pytest.fixture
def cmd_options(request):
    return {
        "tolerance": {"rtol": request.config.getoption("rtol"), "atol": request.config.getoption("atol")},
        "create_mode": request.config.getoption("create_mode"),
    }


@pytest.fixture(scope="module", autouse=True)
def ray_fix():
    import ray
    from agox.utils.ray import reset_ray_pool

    if ray.is_initialized():
        reset_ray_pool()
    yield


def pytest_sessionstart(session):
    from agox.utils.ray import ray_startup

    ray_startup(cpu_count=2, memory=None, tmp_dir=None, include_dashboard=False, max_grace_period=0.0)


test_folder_path = os.path.join(files("agox"), "test/")
test_data_dicts = [
    {"path": "datasets/AgO-dataset.traj", "remove": 6, "name": "AgO"},
    {"path": "datasets/B12-dataset.traj", "remove": 12, "name": "B12"},
    {"path": "datasets/C30-dataset.traj", "remove": 30, "name": "C30"},
]

for dictionary in test_data_dicts:
    dictionary['path'] = os.path.join(test_folder_path, dictionary['path'])

@pytest.fixture(params=test_data_dicts, scope="session")
def dataset_dict(request: pytest.FixtureRequest) -> dict:
    return request.param

@pytest.fixture(params=test_data_dicts, scope="module")
def environment_and_dataset(request: pytest.FixtureRequest) -> Tuple[Environment, List[StandardCandidate]]:
    atoms = read(request.param["path"])
    cell = atoms.get_cell()
    corner = np.array([0, 0, 0])
    remove = request.param["remove"]
    numbers = atoms.get_atomic_numbers()[len(atoms) - remove :]

    template = read(request.param["path"])
    del template[len(template) - remove : len(template)]
    environment = Environment(template=template, numbers=numbers, confinement_cell=cell, confinement_corner=corner)

    data = read(request.param["path"], ":")
    candidates = [StandardCandidate.from_atoms(template, a) for a in data]

    return environment, candidates

@pytest.fixture(scope="session")
def environment_factory(dataset_dict: dict) -> EnvironmentFactory:

    atoms = read(dataset_dict["path"])
    cell = atoms.get_cell()
    corner = np.array([0, 0, 0])
    remove = dataset_dict["remove"]
    numbers = atoms.get_atomic_numbers()[len(atoms) - remove :]

    template = read(dataset_dict["path"])
    del template[len(template) - remove : len(template)]
    environment = Environment(template=template, numbers=numbers, confinement_cell=cell, confinement_corner=corner)

    return EnvironmentFactory(environment)


@pytest.fixture(scope="session")
def dataset_factory(dataset_dict: dict) -> DatasetFactory:
    atoms = read(dataset_dict["path"])
    data = read(dataset_dict["path"], ":")
    candidates = [StandardCandidate.from_atoms(atoms, a) for a in data]
    return DatasetFactory(candidates)



import numpy as np
import pytest

from agox.candidates import CandidateBaseClass
from agox.generators import (
    CenterOfGeometryGenerator,
    RandomGenerator,
    RattleGenerator,
    ReplaceGenerator,
    SamplingGenerator,
)
from agox.helpers.confinement import Confinement

generator_params = [
    (RandomGenerator, (), {}),
    (RandomGenerator, (), {"may_nucleate_at_several_places": True, "c2": 2.5}),
    (RattleGenerator, (), {}),
    (RattleGenerator, (), {"n_rattle": 1}),
    (RattleGenerator, (), {"rattle_amplitude": 1}),
    (ReplaceGenerator, (), {}),
    (ReplaceGenerator, (), {"n_replace": 3}),
    (ReplaceGenerator, (), {"amplitude": 1}),
    (CenterOfGeometryGenerator, (), {}),
    (SamplingGenerator, (), {}),
]


@pytest.fixture(scope="module", params=generator_params)
def candidate(request, sampler, test_environment):
    generator = request.param[0](*request.param[1], **test_environment.get_confinement(), **request.param[2])

    candidate = None
    attempt = 0
    while candidate is None and attempt < 10:
        candidate = generator(sampler, test_environment)[0]
        attempt += 1
    return candidate


def test_output_type(candidate):
    assert isinstance(candidate, CandidateBaseClass)


def test_atomic_formula(candidate, test_environment):
    candidate_numbers = np.sort(np.array(candidate.get_atomic_numbers()))
    env_numbers = np.sort(np.array(test_environment.get_all_numbers()))

    assert (candidate_numbers == env_numbers).all()


def test_confinement(candidate, test_environment):
    indices = test_environment.get_missing_indices()
    positions = candidate.positions[indices, :]
    confinement = test_environment.get_confinement(as_dict=False)
    assert confinement.check_confinement(positions).all()


def test_template(candidate, test_environment):
    template = test_environment.get_template()
    assert (candidate.positions[0 : len(template)] == template.positions).all()


def test_cell(candidate, test_environment):
    assert (candidate.cell == test_environment.get_template().get_cell()).all()



from typing import Callable

import numpy as np
import pytest
from ase import Atoms
from ase.calculators.emt import EMT

from agox.candidates import Candidate
from agox.environments import Environment
from agox.generators import RandomGenerator
from agox.models import CalculatorModel
from agox.samplers.replica_exchange import ReplicaExchangeSampler


@pytest.fixture(scope="module")
def emt_environment() -> Environment:
    template = Atoms(cell=np.eye(3) * 15)
    conf_cell = np.eye(3) * 10
    conf_corner = np.ones(3) * 2.5
    env = Environment(symbols="Cu12", template=template, confinement_cell=conf_cell, confinement_corner=conf_corner)
    return env


@pytest.fixture(scope="module")
def random_generator(emt_environment: Environment) -> RandomGenerator:
    return RandomGenerator(**emt_environment.get_confinement())


@pytest.fixture(scope="module")
def candidate_factory(
    random_generator: RandomGenerator, emt_environment: Environment
) -> Callable[[int], list[Candidate]]:
    def factory(n: int) -> list[Candidate]:
        candidates = [random_generator(sampler=None, environment=emt_environment)[0] for _ in range(n)]
        for i, candidate in enumerate(candidates):
            candidate.add_meta_information("walker_index", i)
            candidate.calc = EMT()

        return candidates

    return factory


@pytest.fixture(scope="function")
def emt_model() -> CalculatorModel:
    return CalculatorModel(calculator=EMT())


@pytest.fixture(scope="function")
def bare_sampler(emt_model: CalculatorModel) -> ReplicaExchangeSampler:
    sampler = ReplicaExchangeSampler(model=emt_model, sample_size=10, t_min=0.2, t_max=2)

    sampler.iteration_counter = 1

    return sampler

@pytest.fixture(scope='function')
def sampler(emt_model: CalculatorModel, candidate_factory: Callable) -> ReplicaExchangeSampler:    
    sampler = ReplicaExchangeSampler(model=emt_model, sample_size=10, 
                                     t_min=0.2, t_max=2)
    
    sampler.iteration_counter = 1
    candidates = candidate_factory(10)
    sampler.setup(candidates)

    return sampler



import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms

from agox import AGOX
from agox.acquisitors import LowerConfidenceBoundAcquisitor
from agox.collectors import ParallelCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator, RattleGenerator
from agox.generators.ce_generator import ComplementaryEnergyGenerator
from agox.generators.complementary_energy.attractor_methods.ce_attractors_current_structure import (
    AttractorCurrentStructure,
)
from agox.generators.complementary_energy.ce_calculators import (
    ComplementaryEnergyDistanceCalculator,
)
from agox.models.descriptors.exponential_density import ExponentialDensity
from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.models.GPR.priors import Repulsive
from agox.postprocessors import ParallelRelaxPostprocess
from agox.samplers import KMeansSampler

# Manually set seed and database-index
seed = 42
database_index = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# database_index = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au8Ni8",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(0)  # From input argument!
database = Database(filename=db_path, order=5)

##############################################################################
# Search Settings:
##############################################################################
descriptor = Fingerprint(environment=environment)
beta = 0.01
k0 = C(beta, (beta, beta)) * RBF()
k1 = C(1 - beta, (1 - beta, 1 - beta)) * RBF()
kernel = C(5000, (1, 1e5)) * (k0 + k1) + Noise(0.01, (0.01, 0.01))
model = GPR(
    descriptor=descriptor,
    kernel=kernel,
    database=database,
    prior=Repulsive(),
    n_optimize=0,  # Set to a multiple of number of cores for a real search.
)


sample_size = 10
sampler = KMeansSampler(descriptor=descriptor, database=database, sample_size=sample_size)

rattle_generator = RattleGenerator(**environment.get_confinement())
random_generator = RandomGenerator(**environment.get_confinement())

lambs = [0.5, 1, 1.5]
rc = 10.0
desc = ExponentialDensity.from_species(["Au", "Ni"], lambs=lambs, rc=rc)
ce_calc = ComplementaryEnergyDistanceCalculator(descriptor=desc)
ce_attractors = AttractorCurrentStructure(desc, attractors_from_template=False)
ce_attractors.attach(database)
ce_generator = ComplementaryEnergyGenerator(ce_calc, desc, ce_attractors, **environment.get_confinement())

# Dict specificies how many candidates are created with and the dict-keys are iterations.
generators = [random_generator, rattle_generator, ce_generator]
num_candidates = {0: [10, 0], 5: [3, 6, 1]}

acquisitor = LowerConfidenceBoundAcquisitor(model=model, kappa=2, order=3)

# CPU-count is set here for Ray - leave it out to use as many cores as are available.
collector = ParallelCollector(
    generators=generators,
    sampler=sampler,
    environment=environment,
    num_candidates=num_candidates,
    order=1,
)

# Number of steps is very low - should be set higher for a real search!
relaxer = ParallelRelaxPostprocess(
    model=acquisitor.get_acquisition_calculator(),
    constraints=environment.get_constraints(),
    order=2,
    start_relax=8,
    optimizer_run_kwargs={"steps": 5},
)

evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "prioritized_candidates"},
    optimizer_kwargs={"logfile": None},
    store_trajectory=True,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 1},
    order=4,
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(collector, acquisitor, relaxer, database, evaluator, seed=seed)

agox.run(N_iterations=10)



from typing import List

import ase.data.colors
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pytest
from ase.constraints import FixAtoms
from matplotlib.collections import PatchCollection
from matplotlib.path import Path

from agox.candidates import StandardCandidate
from agox.utils.plot.plot_atoms import plot_atoms
from agox.utils.plot.utils import plane_to_indices


@pytest.mark.parametrize("plane", ["xy", "xz", "yz"])
@pytest.mark.parametrize("repeat", [0, 1])
def test_plot_atoms(environment_and_dataset, plane, repeat):
    matplotlib.use("Agg")

    _, dataset = environment_and_dataset
    structure: StandardCandidate = dataset[0]

    d0, d1, d2, kv = plane_to_indices(plane)

    sorted_numbers = structure.numbers[np.argsort(kv * structure.positions[:, d2])]

    fig, ax = plt.subplots()
    patch_collection = plot_atoms(ax, structure, plane=plane, repeat=repeat, repeat_patch_kwargs=dict(alpha=0.5))

    expected_repeat = repeat > 0 and structure.get_pbc()[[d0, d1]].sum() > 0

    assert isinstance(patch_collection, PatchCollection)

    assert len(ax.collections) == 1
    assert isinstance(ax.collections[0], PatchCollection)

    p: PatchCollection = ax.collections[0]

    paths: List[Path] = p.get_paths()

    pbc_dimensions = structure.get_pbc()[[d0, d1]].sum()
    expected_copies = (2 * repeat + 1) ** pbc_dimensions

    assert len(paths) == len(structure) * expected_copies

    colors = p.get_facecolor()[:, :3]
    alphas = p.get_facecolor()[:, 3]

    if not expected_repeat:
        np.testing.assert_allclose(colors[: len(structure)], ase.data.colors.jmol_colors[sorted_numbers])
        np.testing.assert_allclose(alphas[: len(structure)], 1)
    else:
        num_original = np.sum(np.isclose(alphas, 1))
        num_repeat = np.sum(np.isclose(alphas, 0.5))

        assert num_original == len(structure)
        assert num_repeat == len(structure) * (expected_copies - 1)

    ax_xmin, ax_xmax = ax.get_xlim()
    ax_ymin, ax_ymax = ax.get_ylim()

    struc_xmin, struc_ymin = structure.positions[:, [d0, d1]].min(axis=0)
    struc_xmax, struc_ymax = structure.positions[:, [d0, d1]].max(axis=0)
    struc_xspan = struc_xmax - struc_xmin
    struc_yspan = struc_ymax - struc_ymin

    assert struc_xmin - struc_xspan < ax_xmin <= struc_xmin
    assert struc_xmax + struc_xspan > ax_xmax >= struc_xmax
    assert struc_ymin - struc_yspan < ax_ymin <= struc_ymin
    assert struc_ymax + struc_yspan > ax_ymax >= struc_ymax

    plt.close(fig)


@pytest.mark.parametrize("plot_constraint", [False, True])
@pytest.mark.parametrize("repeat", [0, 1])
def test_plot_constraint(environment_and_dataset, plot_constraint, repeat):
    matplotlib.use("Agg")

    environment, dataset = environment_and_dataset
    template = environment.get_template()
    structure: StandardCandidate = dataset[0]

    # this part of the test should be changed if `FixAtoms` is more integrated
    # in the environment/candidate
    structure.set_constraint(FixAtoms(indices=range(len(template))))

    fig, ax = plt.subplots()
    patch_collection = plot_atoms(ax, structure, repeat=repeat, plot_constraint=plot_constraint)

    paths: List[Path] = patch_collection.get_paths()

    pbc_dimensions = structure.get_pbc()[[0, 1]].sum()
    expected_copies = (2 * repeat + 1) ** pbc_dimensions

    if plot_constraint:
        # one circle for every atom, one cross for every fixed atom
        expected_n_paths = expected_copies * (len(structure) + len(template))
    else:
        # one circle for every atom
        expected_n_paths = expected_copies * len(structure)

    assert len(paths) == expected_n_paths

    plt.close(fig)



import numpy as np
import pytest
from ase import Atoms
from ase.calculators.lj import LennardJones

from agox.models.descriptors import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.models.GPR.priors import Repulsive
from agox.utils.sparsifiers.CUR import CUR


def model_factory(sparsification_schedule=None):
    # Setup a model:
    descriptor = SOAP.from_species(["H"], periodic=False, r_cut=7)
    kernel = C(1, (1, 1)) * RBF(10, (20, 20))
    sparsifier = CUR(5)

    general_kwargs = dict(
        descriptor=descriptor,
        kernel=kernel,
        sparsifier=sparsifier,
        noise=1,
        prior=Repulsive(ratio=0.7),
        use_ray=False,
        n_optimize=0,
        sparsification_schedule=sparsification_schedule,
    )

    model = SparseGPR(**general_kwargs)
    return model


def pair_potential_data():
    d = np.linspace(0.88, 1.7, 20)
    data = []
    for d_ in d:
        atoms = Atoms("HH", positions=[[0, 0, 0], [0, 0, d_]])
        atoms.calc = LennardJones()
        atoms.get_potential_energy()
        data.append(atoms)

    return data


@pytest.mark.ray
def test_sparse_model_schedule_1():
    # Setup a model:
    model = model_factory()
    data = pair_potential_data()

    # Train the model and it selects sparse points:
    model.train(data)

    # Check that the model has selected the correct number of sparse points:
    assert len(model.Xm) == 5

    # Reset the model and check that it does sparsification:
    model.Xm = model.Xm[0:1, :]
    model.train(data)
    assert len(model.Xm) == 5


@pytest.mark.ray
def test_sparse_model_schedule_2():
    def sparsification_schedule(*args, **kwargs):
        return False

    # Setup a model:
    model = model_factory(sparsification_schedule=sparsification_schedule)

    data = pair_potential_data()

    # Train the model and it selects sparse points:
    model.train(data)

    # Check that the model has selected the correct number of sparse points:
    assert len(model.Xm) == 5

    # Reset the model and check that it doesn't do sparsification:
    model.iteration_counter = 1
    model.Xm = model.Xm[0:1, :]
    model.train(data)
    assert len(model.Xm) == 1



from pathlib import Path

import matplotlib.pyplot as plt

from agox.analysis.plot.property_plot import PropertyPlotter
from agox.analysis.property import EnergyProperty
from agox.analysis.search_data import SearchCollection
from agox.test.test_utils import TemporaryFolder


def test_property_plotter(tmpdir: Path, search_collection: SearchCollection, energy_property: EnergyProperty) -> None:
    plt.switch_backend("agg")

    with TemporaryFolder(tmpdir):
        plotter = PropertyPlotter(search_collection, energy_property)
        fig, ax = plt.subplots()
        plotter.plot(ax)

        plt.savefig("test.png")



import numpy as np
import pytest

from agox.utils.sparsifiers import CUR, MBkmeans, RandomSparsifier


@pytest.mark.parametrize(
    "sparsifier_dict",
    [
        {"sparsifier": RandomSparsifier, "kwargs": {}},
        {"sparsifier": CUR, "kwargs": {}},
        {"sparsifier": MBkmeans, "kwargs": {"exact_points": True}},
    ],
)
def test_sparsifier(sparsifier_dict):
    # Random feautre matix:
    X = np.random.rand(1000, 20)

    # Setup a sparsifier
    sparsifier = sparsifier_dict["sparsifier"](m_points=10, **sparsifier_dict["kwargs"])

    # Sparsify the structures
    Xm, indices = sparsifier.sparsify(X)

    print("Before sparsification: ", X.shape)
    print("After sparsification: ", Xm.shape)

    # Assert that the shapes make sense.
    assert Xm.shape[0] == 10
    assert Xm.shape[1] == 20
    assert len(indices) == 10

    # Assert that using the indices to select rows from X gives the same result as Xm
    assert np.allclose(Xm, X[indices, :])



import numpy as np

from agox.postprocessors.disjoint_filtering import DisjointFilteringPostprocess


def test_disjoint_filtering(environment_and_dataset):
    environment, dataset = environment_and_dataset

    # the dataset includes some structures that would be disjoint by default,
    # so we increase the default scale factor
    postprocessor = DisjointFilteringPostprocess(graph_kwargs=dict(scale_factor=1.6))

    filtered = postprocessor.process_list(dataset)

    assert filtered.count(None) == 0

    # manually make a disjoint version of all structures in the dataset
    for candidate in dataset:
        candidate = candidate.copy()
        candidate.set_pbc(False)
        rattle_idx = np.argmax(candidate.positions[:, 2])
        candidate.positions[rattle_idx, 2] += 5
        assert postprocessor.postprocess(candidate) is None




import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator

# Manually set seed and database-index
seed = 42
database_index = 0

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_cell[2, 2] = 0  # Zero height of confinement cell for the third dimension.
confinement_corner = np.array([3, 3, 6])
environment = Environment(
    template=template,
    symbols="Au4Ni4",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=3)

# We add an additional constraint from ASE to keep atoms confined to 2D during
# relaxations.
# The box constraint will keep them in the specified cell in the other dimensions.
from ase.constraints import FixedPlane

fixed_plane = [FixedPlane(i, [0, 0, 1]) for i in environment.get_missing_indices()]
constraints = environment.get_constraints() + fixed_plane

##############################################################################
# Search Settings:
##############################################################################

random_generator = RandomGenerator(**environment.get_confinement(), environment=environment, order=1)

# Wont relax fully with steps:5 - more realistic setting would be 100+.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    store_trajectory=False,
    order=2,
    constraints=constraints,
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(random_generator, database, evaluator, seed=seed)

agox.run(N_iterations=10)

# Confirm that positions of all atoms placed by the search are in a plane.
# Can leave out of a run.
assert (
    np.array([c.get_positions() for c in database.get_all_candidates()])[:, environment.get_missing_indices(), 2]
    == confinement_corner[2]
).all(), "Not all positions are 2D"



from typing import List

import matplotlib
import matplotlib.pyplot as plt
import pytest
from ase.cell import Cell
from matplotlib.collections import PolyCollection
from matplotlib.path import Path

from agox.utils.plot.plot_cell import plot_cell


@pytest.mark.parametrize("plane", ["xy", "xz", "yz"])
def test_plot_cell(environment_and_dataset, plane):
    matplotlib.use("Agg")

    _, dataset = environment_and_dataset
    cell: Cell = dataset[0].cell

    fig, ax = plt.subplots()
    plot_cell(ax, cell, plane=plane)

    assert len(ax.collections) == 1
    assert isinstance(ax.collections[0], PolyCollection)
    p: PolyCollection = ax.collections[0]

    paths: List[Path] = p.get_paths()
    assert len(paths) == 6

    plt.close(fig)



from ase.calculators.emt import EMT
from ase.calculators.emt import parameters as emt_parameters

from agox.evaluators.local_optimization import LocalOptimizationEvaluator
from agox.evaluators.single_point import SinglePointEvaluator


def test_default_callback(environment_and_dataset):
    environment, dataset = environment_and_dataset

    evaluator = SinglePointEvaluator(EMT())
    for candidate in dataset:
        if any(symbol not in emt_parameters for symbol in candidate.symbols):
            continue

        assert evaluator.evaluate_candidate(candidate) is True


def test_reject_all(environment_and_dataset, capsys):
    environment, dataset = environment_and_dataset

    def reject_all(candidate):
        raise Exception('rejected')

    evaluator = SinglePointEvaluator(EMT(), check_callback=reject_all)
    for candidate in dataset:
        if any(symbol not in emt_parameters for symbol in candidate.symbols):
            continue

        assert evaluator.evaluate_candidate(candidate) is False

        captured = capsys.readouterr()
        assert captured.out == 'Energy calculation failed with exception: rejected\n'
        assert 'Exception: rejected' in captured.err

    assert len(evaluator.evaluated_candidates) == 0


def test_optimization(environment_and_dataset):
    environment, dataset = environment_and_dataset

    def reject_all(candidate):
        raise Exception('rejected')

    evaluator = LocalOptimizationEvaluator(EMT(), check_callback=reject_all)
    for candidate in dataset:
        if any(symbol not in emt_parameters for symbol in candidate.symbols):
            continue

        assert evaluator.evaluate_candidate(candidate) is False

    assert len(evaluator.evaluated_candidates) == 0



import numpy as np

from agox.analysis import SearchData
from agox.analysis.property import ArrayPropertyData, EnergyProperty, FreeEnergyProperty
from agox.utils.thermodynamics.thermodynamics_data import ThermodynamicsData


def free_energy_property() -> FreeEnergyProperty:
    """
    Todo: Not working yet, need to fix it on the GC branch.
    """
    references = {"Ni": 0, "Au": 0}
    chemical_potentials = {"Ni": 0, "Au": 0}

    thermo_data = ThermodynamicsData(references=references, chemical_potentials=chemical_potentials)
    prop = FreeEnergyProperty(thermo_data=thermo_data)
    return prop


def test_energy_property_computes(energy_array_prop: ArrayPropertyData) -> None:
    print(energy_array_prop)


def test_energy_property_hasdata(energy_array_prop: ArrayPropertyData) -> None:
    assert energy_array_prop.data is not None


def test_energy_property_data_nparray(energy_array_prop: ArrayPropertyData) -> None:
    assert isinstance(energy_array_prop.data, np.ndarray)


def test_energy_property_shape(energy_array_prop: ArrayPropertyData) -> None:
    assert energy_array_prop.shape[0] == "Restarts"
    assert energy_array_prop.shape[1] in ["Iterations [#]", "Indices [#]"]


def test_energy_property_axis(energy_array_prop: ArrayPropertyData) -> None:
    assert isinstance(energy_array_prop.axis, tuple)
    assert isinstance(energy_array_prop.axis[0], list)
    assert isinstance(energy_array_prop.axis[1], np.ndarray)


def test_energy_property_name(energy_array_prop: ArrayPropertyData) -> None:
    assert hasattr(energy_array_prop, "name")


def test_energy_property_minimum(energy_array_prop: ArrayPropertyData, search_data: SearchData) -> None:
    energy_property = EnergyProperty()
    min_energy = energy_property.get_minimum([search_data])
    np.testing.assert_allclose(energy_array_prop.data.min(), min_energy)



import os
from importlib import import_module
from pathlib import Path

import numpy as np

from agox.databases import Database
from agox.test.test_utils import TemporaryFolder, check_folder_is_empty, compare_candidates, test_folder_path


def agox_test_run(mode, tmp_path, cmd_options, full_compare=True):
    # Find the path to the current test file
    repo_path = Path(test_folder_path).parent.parent
    env_vars = os.getenv("PYTEST_CURRENT_TEST").split("::")[0]
    test_path = Path(env_vars).parent
    expected_path = repo_path / test_path / "expected_outputs" / f"{mode}_test"

    # Find the script path:
    rel_script_path = test_path / f"script_{mode}.py"
    full_script_path = repo_path / rel_script_path
    script_import = str(rel_script_path).replace("/", ".").replace(".py", "")

    # Check mode & tolerance
    create_mode = cmd_options["create_mode"]
    test_mode = not create_mode
    tolerance = cmd_options["tolerance"]

    if create_mode:
        tmp_path = expected_path
        check_folder_is_empty(tmp_path)

    # Prints for helping later:
    print("====== Test information ======")
    print(f"\tExpected results path: \n\t\t{expected_path}")
    print(f"\tScript path: \n\t\t{full_script_path}")
    print(f"\tScript import: \n\t\t{script_import}")
    print("")

    with TemporaryFolder(tmp_path):
        # This loads the database file from the script file.
        # This means that the documentation can link to this run-file.
        database = import_module(script_import).database

        if test_mode:
            expected_database = Database(f"{expected_path}/db0.db")
            compare_runs(database, expected_database, tolerance, full_compare=full_compare)


def compare_runs(database, expected_database, tolerance, full_compare=True):
    test_candidates = database.get_all_candidates()
    test_energies = database.get_all_energies()

    # Saved database:
    expected_database.restore_to_memory()
    expected_candidates = expected_database.get_all_candidates()
    expected_energies = expected_database.get_all_energies()

    np.testing.assert_allclose(expected_energies, test_energies, **tolerance)

    if full_compare:
        for candidate, expected_candidate in zip(test_candidates, expected_candidates):
            assert compare_candidates(
                candidate, expected_candidate, tolerance
            ), f"Candidates dont match. {candidate.positions = } {expected_candidate.positions = }"

        assert len(expected_candidates) == len(test_candidates), "Different numbers of candidates."

        expected_positions = np.array([atoms.get_positions() for atoms in expected_candidates])
        test_positions = np.array([atoms.get_positions() for atoms in test_candidates])

        np.testing.assert_allclose(expected_energies, test_energies, **tolerance)
        np.testing.assert_allclose(expected_positions, test_positions, **tolerance)



import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase.build import fcc100

from agox import AGOX
from agox.acquisitors import LowerConfidenceBoundAcquisitor
from agox.collectors import ParallelCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator, RattleGenerator
from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.models.GPR.priors import Repulsive
from agox.postprocessors import ParallelRelaxPostprocess
from agox.samplers import KMeansSampler

# Manually set seed and database-index
seed = 41
database_index = 0

##############################################################################
# Calculator
##############################################################################

from ase.calculators.vasp import Vasp

nslots = 8  # Number of cores
calc = Vasp(
    command="mpirun -np {} vasp".format(nslots),
    prec="Low",
    lreal=False,
    lwave=False,
    xc="PBE",
    nwrite=2,
    istart=1,
    icharg=1,
    ispin=1,
    voskown=1,
    lmaxmix=6,
    nelm=500,
    nelmin=4,
    ibrion=-1,
    potim=0.3,
    iwavpr=1,
    nsw=0,
    ialgo=38,
    ismear=0,
    sigma=0.1,
    lorbit=11,
)

##############################################################################
# System & general settings:
##############################################################################

# Define a template surface using ASE functions.
template = fcc100("Au", size=(4, 4, 2), vacuum=5.0)
template.pbc = [True, True, False]
template.positions[:, 2] -= template.positions[:, 2].min()

# Confinement cell matches the template cell in x & y but is smaller in z.
confinement_cell = template.cell.copy()
confinement_cell[2, 2] = 4.0

# Confinement corner is at the origin of the template cell, except in z where
# it is shifted up.
z0 = template.positions[:, 2].max()
confinement_corner = np.array([0, 0, z0])  # Confinement corner is at cell origin

environment = Environment(
    template=template,
    symbols="O4",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
    box_constraint_pbc=[True, True, False],  # Confinement is not periodic in z
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=5)

##############################################################################
# Search Settings:
##############################################################################

# Setup a ML model.
descriptor = Fingerprint(environment=environment)
beta = 0.01
k0 = C(beta, (beta, beta)) * RBF()
k1 = C(1 - beta, (1 - beta, 1 - beta)) * RBF()
kernel = C(5000, (1, 1e5)) * (k0 + k1) + Noise(0.01, (0.01, 0.01))
model = GPR(descriptor=descriptor, kernel=kernel, database=database, prior=Repulsive())

# Sampler to choose candidates to modify using the generators.
sample_size = 10
sampler = KMeansSampler(descriptor=descriptor, database=database, sample_size=sample_size)

# Generators to produce candidates structures
rattle_generator = RattleGenerator(**environment.get_confinement())
random_generator = RandomGenerator(**environment.get_confinement(), contiguous=False)

# Dict specificies how many candidates are created with and the dict-keys are iterations.
generators = [random_generator, rattle_generator]
num_candidates = {0: [10, 0], 5: [3, 7]}

# Collector creates a number of structures in each iteration.
collector = ParallelCollector(
    generators=generators,
    sampler=sampler,
    environment=environment,
    num_candidates=num_candidates,
    order=1,
)

# Acquisitor to choose a candidate to evaluate in the real potential.
acquisitor = LowerConfidenceBoundAcquisitor(model=model, kappa=2, order=3)

# Number of steps is very low - should be set higher for a real search!
relaxer = ParallelRelaxPostprocess(
    model=acquisitor.get_acquisition_calculator(),
    constraints=environment.get_constraints(),
    optimizer_run_kwargs={"steps": 5},
    start_relax=8,
    order=2,
)


# Evaluator to evaluate the candidates in the real potential.
from ase.calculators.calculator import CalculationFailed
from ase.calculators.vasp import Vasp

def callback(candidate):
    if isinstance(candidate.calc, Vasp) and not candidate.calc.read_convergence():
        raise CalculationFailed('VASP: SCF not converged')

evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "prioritized_candidates"},
    optimizer_kwargs={"logfile": None},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 1},
    constraints=environment.get_constraints(),
    store_trajectory=True,
    order=4,
    check_callback=callback,
)

##############################################################################
# Let get the show running!
##############################################################################

# The oder of things here does not matter. But it can be simpler to understand
# what the expected behaviour is if they are put in order.
agox = AGOX(collector, relaxer, acquisitor, evaluator, database, seed=seed)

agox.run(N_iterations=10)



import pytest

from agox.candidates import CandidateBaseClass, Candidate
from agox.generators import (
    CenterOfGeometryGenerator,
    RandomGenerator,
    RattleGenerator,
    ReplaceGenerator,
    SamplingGenerator,
)
from agox.samplers import MetropolisSampler
from agox.test.test_utils import TemporaryFolder

from uuid import uuid4


@pytest.mark.parametrize(
    "generator_class",
    [RandomGenerator, RattleGenerator, ReplaceGenerator, CenterOfGeometryGenerator, SamplingGenerator],
)
class TestGenerator:
    def assertions(self, candidates, environment, sampler):
        for candidate in candidates:
            assert issubclass(candidate.__class__, CandidateBaseClass)
            assert len(candidate) == len(environment.get_all_numbers())
            assert (candidate.cell == environment.get_template().get_cell()).all()

    def setup_generator(self, generator_class, environment, **kwargs):
        return generator_class(environment=environment, **environment.get_confinement(), **kwargs)

    def setup_sampler(self, dataset: list[Candidate]):
        sampler = MetropolisSampler()
        parent = dataset[0]
        parent.add_meta_information('spawn_uuids', [uuid4()])
        sampler.sample = [dataset[0]]
        return sampler

    def test_generators(self, generator_class, environment_and_dataset, tmp_path):
        with TemporaryFolder(tmp_path):
            environment, dataset = environment_and_dataset

            generator = self.setup_generator(generator_class, environment)
            sampler = self.setup_sampler(dataset)
            candidates = [None]
            for i in range(1):
                candidates = generator(sampler, environment)
                if not candidates[0] == None:
                    break
            self.assertions(candidates, environment, sampler)

            assert (tmp_path / f"confinement_plot_{generator.name}.png").is_file()



from pathlib import Path
from typing import List

import pytest
from click.testing import CliRunner

from agox.cli.main import main


def test_analysis_cli_help() -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["analysis", "--help"])
    assert result.exit_code == 0


def test_analysis_cli_basic(database_directories: List[Path]) -> None:
    runner = CliRunner()
    command = ["analysis"]
    for database_directory in database_directories:
        command.append(str(database_directory))

    command.append("--backend")
    command.append("agg")

    result = runner.invoke(main, command)
    assert result.exit_code == 0


@pytest.mark.parametrize("property_name", ["Energy", "energy"])
def test_analysis_cli_energy_property(property_name: str, database_directories: List[Path]) -> None:
    runner = CliRunner()
    command = ["analysis"]
    for database_directory in database_directories:
        command.append(str(database_directory))

    command.append("--backend")
    command.append("agg")

    command.append("--property-name")
    command.append(property_name)

    result = runner.invoke(main, command)
    assert result.exit_code == 0


@pytest.mark.parametrize("flag", ["--success", "--configuration", "--energy", "--distribution"])
def test_analysis_cli_plotting_flags(flag: str, database_directories: List[Path]) -> None:
    runner = CliRunner()
    command = ["analysis"]
    for database_directory in database_directories:
        command.append(str(database_directory))

    command.append("--backend")
    command.append("agg")
    command.append(flag)

    result = runner.invoke(main, command)
    assert result.exit_code == 0


def test_analysis_cli_graph_criteria(database_directories: List[Path]) -> None:
    runner = CliRunner()
    command = ["analysis"]
    for database_directory in database_directories:
        command.append(str(database_directory))

    command.append("--backend")
    command.append("agg")

    command.append("--criterion")
    command.append("graph")

    result = runner.invoke(main, command)
    assert result.exit_code == 0


def test_analysis_cli_eventhandler(mocker, search_collection) -> None:  # noqa
    import matplotlib.pyplot as plt
    import numpy as np

    from agox.analysis.criterion import ThresholdCriterion
    from agox.analysis.property import EnergyProperty
    from agox.analysis.search_analysis import SearchAnalysis
    from agox.cli.cli_analysis import EventHandler

    plt.switch_backend("agg")

    plotting_dict = {
        "configuration": True,
        "energy": True,
        "success": True,
        "distribution": True,
    }

    n_plots = 4
    fig, axes = plt.subplots(1, n_plots, figsize=(5 * n_plots, 5))
    axes = np.atleast_1d(axes)
    axes_dict = {}

    for ax, (label, plot_bool) in zip(axes, plotting_dict.items()):
        if plot_bool:
            axes_dict[label] = ax

    energy_property = EnergyProperty()
    search_analysis = SearchAnalysis(search_collection, energy_property)

    search_analysis.plot_configuration(ax=axes_dict["configuration"], image=0)
    search_analysis.plot_energy(ax=axes_dict["energy"])
    search_analysis.plot_cdf(ax=axes_dict["success"], criterion=ThresholdCriterion(0.1))
    search_analysis.plot_histogram(ax=axes_dict["distribution"], bin_size=0.1, image=0)

    event_handler = EventHandler(search_analysis, axes_dict, fig)

    for key in ["right", "left", "x", "y", "z"]:
        event = mocker.MagicMock()
        event.key = key
        event_handler.on_press(event)



import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RattleGenerator
from agox.samplers import MetropolisSampler

# Manually set seed and database-index
seed = 42
database_index = 0

##############################################################################
# Calculator
##############################################################################

from ase.calculators.orca import ORCA, OrcaProfile

# Replace with path to your orca installation.
profile = OrcaProfile(command="/comm/groupstacks/chemistry/apps/orca/5.0.4/orca")

# For more details on specifying orca settings see: https://wiki.fysik.dtu.dk/ase/ase/calculators/orca.html
calc = ORCA(
    profile=profile,
    directory="orca",
    orcasimpleinput="EnGrad PBE def2-SVP",
    orcablocks="%pal nprocs 8 end",
)  # Note number of processors is set manually!


##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au6",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=4)

##############################################################################
# Search Settings:
##############################################################################

sampler = MetropolisSampler(temperature=0.25, order=3)

rattle_generator = RattleGenerator(**environment.get_confinement(), environment=environment, sampler=sampler, order=1)

# Wont relax fully with steps:5 - more realistic setting would be 100+.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    store_trajectory=False,
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    order=2,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(rattle_generator, database, sampler, evaluator, seed=seed)

agox.run(N_iterations=10)



from pathlib import Path
from typing import List

import pytest

from agox.analysis.property import EnergyProperty
from agox.analysis.search_data import SearchCollection, SearchData


@pytest.fixture(scope="module")
def database_directories() -> List[str]:
    from agox.test.test_utils import test_folder_path

    test_folder_path = Path(test_folder_path)
    database_directories = [
        test_folder_path / "datasets/databases/bh_test_databases/",
        test_folder_path / "datasets/databases/rss_test_databases/",
    ]
    return database_directories


@pytest.fixture(scope="module")
def search_data(database_directories: List[Path]) -> SearchData:
    return SearchData(database_directories[0], reload=True)


@pytest.fixture(scope="module")
def search_collection(database_directories: List[Path]) -> SearchCollection:
    search_collection = SearchCollection(reload=True)
    for directory in database_directories:
        search_collection.add_directory(directory=directory, label=str(directory))

    return search_collection


@pytest.fixture(scope="module", params=["indices", "iterations"])
def energy_property(request: pytest.FixtureRequest) -> EnergyProperty:
    time_axis = request.param
    prop = EnergyProperty(time_axis)
    return prop


@pytest.fixture(scope="module")
def energy_array_prop(search_data: SearchData, energy_property: EnergyProperty) -> None:
    return energy_property.compute(search_data)



from pathlib import Path
from typing import List

import pytest

from agox.analysis.search_data import SearchCollection


@pytest.fixture(scope="module")
def database_directories() -> List[Path]:
    from agox.test.test_utils import test_folder_path

    test_folder_path = Path(test_folder_path)
    database_directories = [
        test_folder_path / "datasets/databases/bh_test_databases/",
        test_folder_path / "datasets/databases/rss_test_databases/",
    ]
    return database_directories


@pytest.fixture(scope="module")
def database_file_path() -> Path:
    from agox.test.test_utils import test_folder_path

    test_folder_path = Path(test_folder_path)
    database_file_path = test_folder_path / "datasets/databases/bh_test_databases/db1.db"

    return database_file_path


@pytest.fixture(scope="module")
def trajectory_file_path() -> Path:
    from agox.test.test_utils import test_folder_path

    test_folder_path = Path(test_folder_path)
    trajectory_file_path = test_folder_path / "datasets/AgO-dataset.traj"

    return trajectory_file_path


@pytest.fixture(scope="module")
def search_collection(database_directories: List[Path]) -> SearchCollection:
    search_collection = SearchCollection(reload=False)
    for directory in database_directories:
        search_collection.add_directory(directory=directory, label=str(directory))

    return search_collection



import pytest

from agox.candidates import CandidateBaseClass
from agox.generators import SymmetryGenerator
from agox.helpers.confinement import Confinement


@pytest.fixture(scope="module")
def random_symmetry_cluster_generator():
    def make_generator(environment):
        generator = SymmetryGenerator(**environment.get_confinement(), sym_type="cluster")
        return generator

    return make_generator


@pytest.fixture(scope="module")
def random_symmetry_cluster_candidate(random_symmetry_cluster_generator, test_environment_non_periodic):
    generator = random_symmetry_cluster_generator(test_environment_non_periodic)

    candidate = None
    attempt = 0
    while candidate is None and attempt < 100:  # This is bad, but...
        generator_output = generator(None, test_environment_non_periodic)
        attempt += 1

        if len(generator_output) > 0:
            candidate = generator_output[0]

    return candidate


def test_output_type(random_symmetry_cluster_candidate):
    assert isinstance(random_symmetry_cluster_candidate, CandidateBaseClass)


def test_output_symmetry_type(random_symmetry_cluster_candidate):
    space_group = random_symmetry_cluster_candidate.get_meta_information("space_group")
    assert isinstance(space_group, str)


def test_confinement(random_symmetry_cluster_candidate, test_environment_non_periodic):
    indices = test_environment_non_periodic.get_missing_indices()
    positions = random_symmetry_cluster_candidate.positions[indices, :]
    confinement = test_environment_non_periodic.get_confinement(as_dict=False)
    assert confinement.check_confinement(positions).all()


def test_template(random_symmetry_cluster_candidate, test_environment_non_periodic):
    template = test_environment_non_periodic.get_template()
    assert (random_symmetry_cluster_candidate.positions[0 : len(template)] == template.positions).all()


def test_cell(random_symmetry_cluster_candidate, test_environment_non_periodic):
    assert (random_symmetry_cluster_candidate.cell == test_environment_non_periodic.get_template().get_cell()).all()



import pytest

from agox.test.run_tests.run_utils import agox_test_run


@pytest.mark.skip(reason="Some issue with this test currently (30-1-2025), but I want the others to run in CI.")
@pytest.mark.ray
def test_graph_filtering_gofee(tmp_path, cmd_options):
    mode = "graph_filtering_gofee"  # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options)



import pytest

from agox.test.run_tests.run_utils import agox_test_run


@pytest.mark.ray
def test_lgpr_bh(tmp_path, cmd_options):
    # pytest.skip('This test is not yet working always due to tolerances on different machines.')
    mode = "lgpr_bh"  # This determines the script that is imported.
    options = {"tolerance": {"rtol": 1e-03, "atol": 1e-03}, "create_mode": cmd_options["create_mode"]}
    agox_test_run(mode, tmp_path, options)



import pytest

from agox.test.run_tests.run_utils import agox_test_run


@pytest.mark.ray
def test_gofee_slab_symmetry(tmp_path, cmd_options, ray_fix):
    mode = "gofee_slab_symmetry"  # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options, full_compare=False)



import numpy as np



def model_tester(
    model_maker,
    model_args,
    model_kwargs,
    data,
    test_mode=True,
    expected_data=None,
    tolerance=None,
    force_tolerance=None,
):
    seed = 42
    np.random.seed(seed)

    # Data split:
    training_data = data[0 : len(data) // 2]
    training_energies = np.array([atoms.get_potential_energy() for atoms in training_data])
    test_data = data[len(data) // 2 :]

    # Make model instance:
    model = model_maker(*model_args, **model_kwargs)

    # Train the model:
    model.train(training_data)

    # Test on the remaining data:
    E = np.zeros(len(test_data))
    F = []
    for i, test_atoms in enumerate(test_data):
        test_atoms.calc = model
        E[i] = test_atoms.get_potential_energy()
        F += list(test_atoms.get_forces().flatten())
    F = np.array(F)

    if test_mode:
        expected_energies = expected_data["E"]
        np.testing.assert_allclose(E, expected_energies, **tolerance)
        expected_forces = expected_data["F"]

        if force_tolerance is None:
            force_tolerance = tolerance

        np.testing.assert_allclose(F, expected_forces, **force_tolerance)

        # # Model parameters:
        # parameters = model.get_model_parameters()
        # recreated_model = model_maker(*model_args, **model_kwargs)

        # recreated_model.set_model_parameters(parameters)
        # recreated_energies = np.array([recreated_model.predict_energy(atoms) for atoms in test_data])
        # np.testing.assert_allclose(E, recreated_energies)

    return {"E": E, "F": F}



import numpy as np
import pytest

from agox.models.descriptors import SOAP
from agox.test.test_utils import (
    get_test_data,
    get_test_environment,
    load_expected_data,
    save_expected_data,
    test_data_dicts,
    test_folder_path,
)


@pytest.mark.parametrize("test_data_dict", test_data_dicts)
def test_descriptor(test_data_dict, cmd_options):
    create_mode = cmd_options["create_mode"]
    tolerance = cmd_options["tolerance"]

    path = test_data_dict["path"]
    remove = test_data_dict["remove"]
    name = test_data_dict["name"]
    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)

    species = set(data[0].get_chemical_symbols())
    descriptor = SOAP.from_species(species)

    F = np.array(descriptor.get_features(data))

    name = f"{test_folder_path}/descriptor_tests/expected_outputs/{descriptor.name}_{name}.pckl"

    if create_mode:
        save_expected_data(name, F)
    else:
        F_exp = load_expected_data(name)
        np.testing.assert_allclose(F, F_exp, **tolerance)



from agox.candidates import Candidate
from agox.samplers.replica_exchange import ReplicaExchangeSampler


def test_bare_sampler_type(bare_sampler: ReplicaExchangeSampler) -> None:
    assert isinstance(bare_sampler, ReplicaExchangeSampler)


def test_bare_sampler_is_empty(bare_sampler: ReplicaExchangeSampler) -> None:
    for sample_member in bare_sampler.sample:
        assert sample_member.empty()


def test_sampler_not_empty(sampler: ReplicaExchangeSampler) -> None:
    for sample_member in sampler.sample:
        assert not sample_member.empty()


def test_sampler_get_walker_type(sampler: ReplicaExchangeSampler) -> None:
    for i in range(sampler.sample_size):
        walker = sampler.get_walker(i)
        assert isinstance(walker, Candidate)


def test_sampler_members_have_calculators(sampler: ReplicaExchangeSampler) -> None:
    for walker in sampler.sample:
        if not walker.empty():
            assert walker.candidate.calc is not None


def test_sampler_energy_match(sampler: ReplicaExchangeSampler) -> None:
    for walker in sampler.sample:
        assert walker.get_energy() == walker.candidate.get_potential_energy()



from typing import Callable

import pytest

from agox.candidates import Candidate
from agox.samplers.replica_exchange import ReplicaExchangeSample
from agox.samplers.replica_exchange.sample import SampleMember


@pytest.fixture(scope='function')
def empty_sample() -> ReplicaExchangeSample:
    return ReplicaExchangeSample(10)

@pytest.fixture(scope='function')
def sample(candidate_factory: Callable[[int], list[Candidate]]) -> ReplicaExchangeSample:
    sample_size = 10
    sample = ReplicaExchangeSample(sample_size)
    candidates = candidate_factory(sample_size)
    for i, candidate in enumerate(candidates):
        sample.update(i, candidate, candidate.get_potential_energy())
    return sample

def test_sample(sample: ReplicaExchangeSample) -> None:
    assert isinstance(sample, ReplicaExchangeSample)

def test_sample_getitem(sample: ReplicaExchangeSample) -> None:
    assert isinstance(sample[0], SampleMember)

def test_sample_setitem(sample: ReplicaExchangeSample, candidate_factory: Callable) -> None:
    sample[0] = candidate_factory(1)[0]
    assert isinstance(sample[0], SampleMember)

def test_sample_swap(sample: ReplicaExchangeSample) -> None:
    e0 = sample[0].get_energy()
    e1 = sample[1].get_energy()
    c0 = sample[0].candidate
    c1 = sample[1].candidate

    sample.swap(0, 1)
    assert sample[0].get_energy() == e1
    assert sample[1].get_energy() == e0

    assert sample[0].candidate == c1
    assert sample[1].candidate == c0
    assert sample[0] != c0
    assert sample[1] != c1

def test_empty_sample_is_empty(empty_sample: ReplicaExchangeSample) -> None:
    assert all([sample.empty() for sample in empty_sample.sample])

def test_sample_is_not_empty(sample: ReplicaExchangeSample) -> None:
    assert not all([sample.empty() for sample in sample.sample])

def test_sample_update(sample: ReplicaExchangeSample, candidate_factory: Callable) -> None:

    initial_energies = [sample[i].get_energy() for i in range(len(sample.sample))]

    candidates = candidate_factory(10)
    new_energies_unset = [candidate.get_potential_energy() for candidate in candidates]

    for i, candidate in enumerate(candidates):
        sample.update(i, candidate, candidate.get_potential_energy())

    new_energies_set = [sample[i].get_energy() for i in range(len(sample.sample))]

    assert initial_energies != new_energies_unset
    assert new_energies_unset == new_energies_set

def test_sample_len(sample: ReplicaExchangeSample) -> None:
    assert len(sample) == len(sample.sample)

def test_sample_iter(sample: ReplicaExchangeSample) -> None:
    for sample_member in sample:
        assert isinstance(sample_member, SampleMember)

    


import numpy as np
from ase.atom import Atom
from ase.atoms import Atoms
from ase.build import fcc111
from numpy.typing import NDArray

from agox.candidates.standard import StandardCandidate
from agox.postprocessors.surface_centering import SurfaceCenteringPostprocess


def test_surface_centering():
    rng = np.random.default_rng(1234)

    size = (5, 5, 1)
    template: Atoms = fcc111("Cu", size)
    unit_cell: NDArray = template.info["adsorbate_info"]["cell"]

    p = SurfaceCenteringPostprocess(template_size=size)

    for _ in range(10):
        # generate a random position within the minimal unit cell
        uc_position = rng.random(2)
        xy_position = uc_position @ unit_cell

        # add an atom at this position
        atom = Atom("Zn", position=[xy_position[0], xy_position[1], 2.0])
        candidate = StandardCandidate.from_atoms(template=template, atoms=template + atom)

        # center the atom
        centered = p.postprocess(candidate)

        # get the xy offset applied to the atom
        dx = (centered[-1].position - atom.position)[:2]

        # get the offset in terms of minimal unit cell units
        duc = np.linalg.solve(unit_cell.T, dx)

        # check whether this offset is integral
        assert np.all(np.isclose(duc, np.round(duc)))



import pytest

from agox.test.run_tests.run_utils import agox_test_run


@pytest.mark.parametrize("mode", ["bh_cluster_emt", "bh_surface_emt", "bh_bulk_emt"])
def test_bh(mode, tmp_path, cmd_options):
    agox_test_run(mode, tmp_path, cmd_options)



import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms
from ase.optimize import BFGS

from agox import AGOX
from agox.acquisitors import MetaInformationAcquisitor
from agox.collectors import ReplicaExchangeCollector
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators import RandomGenerator, RattleGenerator
from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR import GPR
from agox.models.GPR.kernels import RBF, Noise
from agox.models.GPR.kernels import Constant as C
from agox.models.GPR.priors import Repulsive
from agox.postprocessors import ParallelRelaxPostprocess
from agox.samplers import ReplicaExchangeSampler

seed = 42
database_index = 0

sample_size = 10
rattle_amplitudes = np.linspace(0.1, 5, sample_size)

################################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

db_path = "db{}.db".format(database_index)
database = Database(filename=db_path, order=6)

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au12Ni2",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

##############################################################################
# Search Settings:
##############################################################################

# Setup a ML model.
descriptor = Fingerprint(environment=environment)
beta = 0.01
k0 = C(beta, (beta, beta)) * RBF()
k1 = C(1 - beta, (1 - beta, 1 - beta)) * RBF()
kernel = C(5000, (1, 1e5)) * (k0 + k1) + Noise(0.01, (0.01, 0.01))
model = GPR(descriptor=descriptor, kernel=kernel, database=database, prior=Repulsive())

sampler = ReplicaExchangeSampler(
    t_min=0.02, t_max=2, swap="down", sample_size=sample_size, order=3, model=model
)


# Setup the generators to be used, as a dictionary
generator_dict = {}
random_generator = RandomGenerator(contiguous=True,**environment.get_confinement())
generator_dict['random'] = random_generator
for i, t in enumerate(sampler.temperatures):
    generator_dict[t] = RattleGenerator(
        **environment.get_confinement(), 
        rattle_amplitude=rattle_amplitudes[i],
        n_rattle=len(environment.get_missing_indices())
        )

collector = ReplicaExchangeCollector.from_sampler(
    sampler, environment, rattle_amplitudes)

# Allow the amplitudes to vary over the course of the calculation to normalise acceptance probability
for key, generator in generator_dict.items(): 
    collector.add_generator_update(key,generator,"rattle_amplitude")

# Local optimisation parameters
relaxer = ParallelRelaxPostprocess(
    model=model,
    constraints=environment.get_constraints(),
    optimizer_run_kwargs={"steps": 5, "fmax": 0.1},
    start_relax=0,
    optimizer=BFGS,
    order=2,
)

# With skip function we can choose conditions under which we will skip the evaluation step. 
# In this case, we only perform evaluation if the iteration number is odd.
def skip_function(self, candidate_list): 
    if  self.get_iteration_counter() % 2 != 0: 
        return False
    else: 
        return True

# We select by walker index i.e. those walkers with lowest index are selected first for evaluation.
acquisitor = MetaInformationAcquisitor(
    meta_key="walker_index", order=4, skip_function=skip_function
)

# Each iteration the lowest temperature walker/replica will be evaluated.
evaluator = LocalOptimizationEvaluator(
    calc,
    number_to_evaluate=1,
    optimizer_kwargs={"logfile": None},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 0},
    constraints=environment.get_constraints(),
    order=5,
)

agox = AGOX(
    collector,
    relaxer,
    sampler,
    acquisitor,
    evaluator,
    database,
    seed=seed,
)

# ##############################################################################
# # Lets get the show running!
# ##############################################################################

agox.run(N_iterations=10)



import numpy as np
import pytest

from agox.models.descriptors import SOAP
from agox.models.GPR import SparseGPR
from agox.models.GPR.kernels import RBF
from agox.models.GPR.kernels import Constant as C
from agox.test.test_utils import get_test_data, get_test_environment, test_data_dicts


@pytest.mark.ray
@pytest.mark.parametrize("test_data_dict", test_data_dicts)
def test_ray_global_gpr(test_data_dict, ray_fix, cmd_options):
    # Tolerances:
    tolerance = cmd_options["tolerance"]

    path = test_data_dict["path"]
    remove = test_data_dict["remove"]

    environment = get_test_environment(path, remove)
    data = get_test_data(path, environment)

    train_data = data[0:25]
    test_data = data[25:35]

    # Make a global model:
    descriptor = SOAP(environment=environment)

    # If the kernel hyperparameters are allowed to take extreme values (large amplitude, small length scale),
    # it can ocassionally fail - even when the hyperparameters are properly transferred to the actors.
    # But I think that has to be a numerical issue?
    kernel = C(1, (1, 100)) * RBF(10, (10, 50))

    model = SparseGPR(
        descriptor=descriptor, kernel=kernel, n_optimize=2, use_ray=True, unc_jitter=1e-3, train_uncertainty=True
    )

    model.train(train_data)
    model.pool.synchronize_module(model)

    # Check properties on actors:
    result_dicts = model.pool.get_module_attributes(model, model.dynamic_attributes)

    for result_dict in result_dicts:
        for attribute_name in ["X", "Xm", "Kmm_inv", "alpha", "C_inv"]:
            np.testing.assert_allclose(result_dict[attribute_name], model.__getattribute__(attribute_name), **tolerance)

        # Kernel check seperately because it is annoying.
        np.testing.assert_allclose(result_dict["kernel"].theta, model.kernel.theta, **tolerance)

    # Predictions without Ray:
    energies = np.array(model.predict_energy(test_data))
    forces = np.array(model.predict_forces(test_data))
    uncertainties = np.array(model.predict_uncertainty(test_data))
    unceertainty_forces = np.array(model.predict_uncertainty_forces(test_data))

    # Predictions with Ray:
    def predict(model, atoms):
        E = np.array(model.predict_energy(atoms))
        F = np.array(model.predict_forces(atoms))
        sigma = np.array(model.predict_uncertainty(atoms))
        sigma_F = np.array(model.predict_uncertainty_forces(atoms))

        results = {"E": E, "F": F, "sigma": sigma, "sigma_F": sigma_F}
        return results

    pool = model.pool
    # Make predictions on all actors:
    key = pool.get_key(model)
    results = pool.execute_on_actors(predict, [key], [test_data], {})

    # Check that all predictions on every actor are the same:
    results_actor_0 = results[0]
    for key in results[0].keys():
        for results_actor in results[1:]:
            np.testing.assert_allclose(results_actor_0[key], results_actor[key])

    # # Check that predictions are the same as without Ray:
    np.testing.assert_allclose(energies, results_actor_0["E"], **tolerance)
    np.testing.assert_allclose(forces, results_actor_0["F"], **tolerance)
    np.testing.assert_allclose(uncertainties, results_actor_0["sigma"], **tolerance)
    np.testing.assert_allclose(unceertainty_forces, results_actor_0["sigma_F"], **tolerance)



from pathlib import Path


from agox.test.test_utils import test_folder_path

test_folder_path = Path(test_folder_path)
database_directories = {"bh": test_folder_path / "datasets/databases/bh_test_databases/"}


def test_restart_data():
    import numpy as np

    from agox.analysis.search_data import RestartData, read_database
    from agox.databases import Database

    path = database_directories["bh"] / "db1.db"
    _, candidates = read_database(path, Database)
    restart_data = RestartData(path, candidates)

    assert len(restart_data) == len(candidates)
    assert hasattr(restart_data, "identifier")
    assert isinstance(restart_data.energies, np.ndarray)



import matplotlib

matplotlib.use("Agg")

import numpy as np
from ase import Atoms

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators.block_generators import RattleBlockGenerator
from agox.samplers import MetropolisSampler

# Manually set seed and database-index
seed = 42
database_index = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# database_index = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au2Ni2",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

sampler = MetropolisSampler(temperature=0.25, order=3)


building_block = Atoms("AuNi", positions=[[0, 0, 0], [2.4, 0, 0]])

rattle_generator = RattleBlockGenerator(
    **environment.get_confinement(),
    building_blocks=[building_block],
    N_blocks=[2],
    environment=environment,
    sampler=sampler,
    order=1,
)

# Wont relax fully with steps:5 - more realistic setting would be 100+.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    store_trajectory=False,
    order=2,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(rattle_generator, database, sampler, evaluator, seed=seed)

agox.run(N_iterations=10)



import numpy as np
from itertools import product
from ase.ga.utilities import closest_distances_generator
cimport numpy as np

from libc.math cimport *

from cymem.cymem cimport Pool
cimport cython

# Custom functions
ctypedef struct Point:
    double coord[3]

cdef Point subtract(Point p1, Point p2):
    cdef Point p
    p.coord[0] = p1.coord[0] - p2.coord[0]
    p.coord[1] = p1.coord[1] - p2.coord[1]
    p.coord[2] = p1.coord[2] - p2.coord[2]
    return p

cdef Point add(Point p1, Point p2):
    cdef Point p
    p.coord[0] = p1.coord[0] + p2.coord[0]
    p.coord[1] = p1.coord[1] + p2.coord[1]
    p.coord[2] = p1.coord[2] + p2.coord[2]
    return p

cdef double norm(Point p):
    return sqrt(p.coord[0]*p.coord[0] + p.coord[1]*p.coord[1] + p.coord[2]*p.coord[2])

cdef double euclidean(Point p1, Point p2):
    return norm(subtract(p1,p2))

def convert_atom_types(num):
    cdef int Natoms = len(num)
    cdef list atomic_types = sorted(list(set(num)))
    cdef int Ntypes = len(atomic_types)
    cdef list num_converted = [0]*Natoms
    cdef int i, j
    for i in range(Natoms):
        for j in range(Ntypes):
            if num[i] == atomic_types[j]:
                num_converted[i] = j
    return num_converted

class Repulsive():

    def __init__(self, rcut=3., ratio=0.7):
        self.rcut = rcut
        self.ratio = ratio
        self.cell = None

    def predict_energy(self, a):
        if self.cell is None:
            self.prepare(a)
        if not np.array_equal(self.cell, a.get_cell()):
            self.prepare(a)


        # Memory allocation pool
        cdef Pool mem
        mem = Pool()

        cdef int Natoms = len(a)
        cdef double rcut = self.rcut

        cdef list x_np = a.get_positions().tolist()
        cdef Point *x
        x = <Point*>mem.alloc(Natoms, sizeof(Point))
        cdef int m
        for m in range(Natoms):
            x[m].coord[0] = x_np[m][0]
            x[m].coord[1] = x_np[m][1]
            x[m].coord[2] = x_np[m][2]

        # Get neighbourcells and convert to Point-struct
        cdef int Ncells = self.Ncells
        cdef list cell_displacements_old = self.cell_displacements
        cdef Point *cell_displacements
        cell_displacements = <Point*>mem.alloc(Ncells, sizeof(Point))
        for m in range(Ncells):
            cell_displacements[m].coord[0] = cell_displacements_old[m][0]
            cell_displacements[m].coord[1] = cell_displacements_old[m][1]
            cell_displacements[m].coord[2] = cell_displacements_old[m][2]

        # Convert 2body bondtype list into c-array
        cdef int Ntypes = self.Ntypes
        cdef list blmin_old = self.blmin
        cdef double *blmin
        blmin = <double*>mem.alloc(Ntypes*Ntypes, sizeof(double))
        for m in range(Ntypes*Ntypes):
            blmin[m] = blmin_old[m]

        # Get converted atom Ntypes
        cdef list num_converted_old = convert_atom_types(a.get_atomic_numbers())
        cdef int *num_converted
        num_converted = <int*>mem.alloc(Natoms, sizeof(int))
        for m in range(Natoms):
            num_converted[m] = num_converted_old[m]

        cdef double E=0, rmin, radd
        cdef int i, j
        for i in range(Natoms):
            xi = x[i]
            for cell_index in range(Ncells):
                displacement = cell_displacements[cell_index]
                for j in range(Natoms):
                    rmin = blmin[num_converted[i]*Ntypes + num_converted[j]]
                    radd = 1 - rmin
                    xj = add(x[j], displacement)
                    r = euclidean(xi, xj)
                    if r > 1e-5 and r < rcut:
                        r_scaled = r + radd
                        E += 1/pow(r_scaled,12)
        return E/2

    def predict_forces(self, a):
        # Memory allocation pool
        cdef Pool mem
        mem = Pool()

        cdef int Natoms = len(a)
        cdef int dim = 3
        cdef double rcut = self.rcut

        cdef list x_np = a.get_positions().tolist()
        cdef Point *x
        x = <Point*>mem.alloc(Natoms, sizeof(Point))
        cdef int m
        for m in range(Natoms):
            x[m].coord[0] = x_np[m][0]
            x[m].coord[1] = x_np[m][1]
            x[m].coord[2] = x_np[m][2]

        # Get neighbourcells and convert to Point-struct
        cdef int Ncells = self.Ncells
        cdef list cell_displacements_old = self.cell_displacements
        cdef Point *cell_displacements
        cell_displacements = <Point*>mem.alloc(Ncells, sizeof(Point))
        for m in range(Ncells):
            cell_displacements[m].coord[0] = cell_displacements_old[m][0]
            cell_displacements[m].coord[1] = cell_displacements_old[m][1]
            cell_displacements[m].coord[2] = cell_displacements_old[m][2]

        # Convert 2body bondtype list into c-array
        cdef int Ntypes = self.Ntypes
        cdef list blmin_old = self.blmin
        cdef double *blmin
        blmin = <double*>mem.alloc(Ntypes*Ntypes, sizeof(double))
        for m in range(Ntypes*Ntypes):
            blmin[m] = blmin_old[m]

        # Get converted atom Ntypes
        cdef list num_converted_old = convert_atom_types(a.get_atomic_numbers())
        cdef int *num_converted
        num_converted = <int*>mem.alloc(Natoms, sizeof(int))
        for m in range(Natoms):
            num_converted[m] = num_converted_old[m]

        #cdef double rmin = 0.7 * self.cov_dist
        #cdef double radd = 1 - rmin

        # Initialize Force object
        cdef double *dE
        dE = <double*>mem.alloc(Natoms * dim, sizeof(double))

        cdef int i, j, k
        for i in range(Natoms):
            xi = x[i]
            for cell_index in range(Ncells):
                displacement = cell_displacements[cell_index]
                for j in range(Natoms):
                    rmin = blmin[num_converted[i]*Ntypes + num_converted[j]]
                    radd = 1 - rmin
                    xj = add(x[j], displacement)
                    r = euclidean(xi,xj)
                    if r > 1e-5 and r < rcut:
                        r_scaled = r + radd
                        rijVec = subtract(xi,xj)

                        for k in range(dim):
                            dE[dim*i + k] += -12*rijVec.coord[k] / (pow(r_scaled,13)*r)
                            dE[dim*j + k] += 12*rijVec.coord[k] / (pow(r_scaled,13)*r)

        dE_np = np.zeros((Natoms, dim))
        for i in range(Natoms):
            for k in range(dim):
                dE_np[i,k] = dE[dim*i + k]

        return - dE_np.reshape(-1, 3)/2

    def __get_neighbour_cells_displacement(self, pbc, cell):
        # Calculate neighbour cells
        cell_vec_norms = np.linalg.norm(cell, axis=0)
        neighbours = []
        for i in range(3):
            if pbc[i]:
                ncellmax = int(np.ceil(abs(self.rcut/cell_vec_norms[i]))) + 1  # +1 because atoms can be outside unitcell.
                neighbours.append(range(-ncellmax,ncellmax+1))
            else:
                neighbours.append([0])

        neighbourcells_disp = []
        for x,y,z in product(*neighbours):
            xyz = (x,y,z)
            displacement = np.dot(xyz, cell)
            neighbourcells_disp.append(list(displacement))

        return neighbourcells_disp


    def prepare(self, a):
        self.pbc = a.get_pbc()
        self.cell = a.get_cell()

        self.cell_displacements = self.__get_neighbour_cells_displacement(self.pbc, self.cell)
        self.Ncells = len(self.cell_displacements)

        num = a.get_atomic_numbers()
        atomic_types = sorted(list(set(num)))
        self.Ntypes = len(atomic_types)

        blmin_dict = closest_distances_generator(num, ratio_of_covalent_radii=self.ratio)

        self.blmin = -np.ones((self.Ntypes, self.Ntypes))
        for i1, type1 in enumerate(atomic_types):
            for i2, type2 in enumerate(atomic_types):
                self.blmin[i1,i2] = blmin_dict[(type1, type2)]
        self.blmin = self.blmin.reshape(-1).tolist()



import os
import pickle
from importlib.resources import files
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union
from uuid import uuid4

import numpy as np
from ase import Atoms
from ase.io import read

from agox.candidates import StandardCandidate
from agox.environments import Environment
from agox.samplers import SamplerBaseClass

test_folder_path = os.path.join(files("agox"), "test/")

test_data_dicts = [
    {"path": "datasets/AgO-dataset.traj", "remove": 6, "name": "AgO"},
    {"path": "datasets/B12-dataset.traj", "remove": 12, "name": "B12"},
    {"path": "datasets/C30-dataset.traj", "remove": 30, "name": "C30"},
]

for dictionary in test_data_dicts:
    dictionary["path"] = os.path.join(test_folder_path, dictionary["path"])


class TemporaryFolder:
    def __init__(self, path: Union[str, Path]) -> None:
        d = path / ""
        if not os.path.exists(d):
            d.mkdir()
        self.start_dir = os.getcwd()
        self.path = d

    def __enter__(self) -> None:
        os.chdir(self.path)

    def __exit__(self, *args) -> None:
        os.chdir(self.start_dir)


def compare_candidates(atoms_1: Atoms, atoms_2: Atoms, tolerance: float) -> bool:
    if atoms_1 is None and atoms_2 is None:
        return True

    np.testing.assert_allclose(atoms_1.positions, atoms_2.positions, **tolerance)
    np.testing.assert_allclose(atoms_1.cell, atoms_2.cell, **tolerance)
    np.testing.assert_allclose(atoms_1.numbers, atoms_2.numbers, **tolerance)
    return True


def get_test_environment(path: Union[str, Path], remove: int) -> Environment:
    from agox.environments import Environment

    atoms = read(path, index=0)
    cell = atoms.get_cell()
    corner = np.array([0, 0, 0])
    numbers = atoms.get_atomic_numbers()[len(atoms) - remove :]

    template = atoms.copy()
    del template[len(template) - remove : len(template)]
    environment = Environment(template=template, numbers=numbers, confinement_cell=cell, confinement_corner=corner)

    return environment


def get_test_data(path: Union[str, Path], environment: Environment) -> List[StandardCandidate]:
    template = environment.get_template()
    atoms_data = read(path, index=":")
    candidate_data = [StandardCandidate.from_atoms(template, atoms) for atoms in atoms_data]
    return candidate_data


def get_test_sampler(data: List[StandardCandidate]) -> SamplerBaseClass:
    from agox.samplers import SamplerBaseClass

    class DummySampler(SamplerBaseClass):
        name = "DummySampler"
        initialized = False

        def __init__(self, data: List[StandardCandidate]) -> None:
            self.sample = data

            for candidate in self.sample:
                candidate.add_meta_information("spawn_uuids", [uuid4()])

        def setup(self, *args, **kwargs):
            return None

    return DummySampler(data)


def get_name(module_name: str, subfolder: str, dataset_name: str, parameter_index: str) -> str:
    folder = "expected_outputs/"
    name = f"{test_folder_path}{subfolder}{folder}{module_name}_data{dataset_name}_parameter{parameter_index}.pckl"
    return name


def save_expected_data(name: str, data: Any) -> None:
    check_file_is_deleted(name)
    with open(name, "wb") as f:
        pickle.dump(data, f)


def load_expected_data(name: str) -> Any:
    with open(name, "rb") as f:
        data = pickle.load(f)
    return data


def label_dict_list(list_of_dicts: List[Dict]) -> List[Tuple[Dict, int]]:
    for index, dictionary in enumerate(list_of_dicts):
        list_of_dicts[index] = (dictionary, index)
    return list_of_dicts


def check_folder_is_empty(path: Union[str, Path]) -> None:
    path = Path(path)

    if not path.exists():
        print("Creating path")
        path.mkdir(parents=True)

    files = list(path.glob("*"))
    assert len(files) == 0, "Expected path is not empty, delete the files if you are sure you want to remake them!"


def check_file_is_deleted(filepath: Path) -> None:
    assert not os.path.exists(
        filepath
    ), "File already exists - manually delete it if youre sure you want to recreate it."
